/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import javax.vecmath.Matrix4f;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Representation of a purely ambient light source with no other abilities.
 * <p>
 *
 * An ambient light has no direction or location, but allows a base colour
 * to be added to all objects within it's influence. This class allows for the
 * basic lighting abilities to be used, without needing to drag in any other
 * sort of light source like the other derived light types do.
 *
 * @author Justin Couch
 * @version $Revision: 1.7 $
 */
public class AmbientLight extends Light
{
    /**
     * Creates a light with the colour set to black.
     */
    public AmbientLight()
    {
        super(AMBIENT_TYPE);

    }

    /**
     * Create a light with the given base colour.Colour must be in the range
     * [0, 1] otherwise an exception is generated.
     *
     * @param col The new colour value to use
     * @throws IllegalArgumentException The colour value is out of range
     */
    public AmbientLight(float[] col)
        throws IllegalArgumentException
    {
        super(AMBIENT_TYPE);

        if(col[0] < 0 || col[0] > 1)
            throw new IllegalArgumentException("Red component out of range");

        if(col[1] < 0 || col[1] > 1)
            throw new IllegalArgumentException("Green component out of range");

        if(col[2] < 0 || col[2] > 1)
            throw new IllegalArgumentException("Blue component out of range");

        ambientColor[0] = col[0];
        ambientColor[1] = col[1];
        ambientColor[2] = col[2];
    }

    //---------------------------------------------------------------
    // Methods defined by SceneGraphObject
    //---------------------------------------------------------------

    /**
     * Overloaded form of the render() method to render the light details given
     * the specific Light ID used by OpenGL. Since the active light ID for this
     * node may vary over time, a fixed ID cannot be used by OpenGL. The
     * renderer will always call this method rather than the normal render()
     * method. The normal post render will still be called
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     * @param lightId the ID of the light to make GL calls with
     */
    public void render(GL gl, GLU glu, int lightId)
    {
        gl.glLightfv(lightId, GL.GL_SPECULAR, specularColor);
        gl.glLightfv(lightId, GL.GL_DIFFUSE, diffuseColor);
        gl.glLightfv(lightId, GL.GL_AMBIENT, ambientColor);

        gl.glEnable(lightId);
    }

    /*
     * Overloaded form of the postRender() method to render the light details given
     * the specific Light ID used by OpenGL. Since the active light ID for this
     * node may vary over time, a fixed ID cannot be used by OpenGL. The
     * renderer will always call this method rather than the normal postRender()
     * method. The normal post render will still be called
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     * @param lightId the ID of the light to make GL calls with
     */
    public void postRender(GL gl, GLU glu, int lightId)
    {
        gl.glDisable(lightId);
    }

    /**
     * This method is called to render this node. All openGL commands needed
     * to render the node should be executed.  Any transformations needed
     * should be added to the transformation stack premultiplied.  This
     * method must be re-entrant as it can be called from multiple
     * places at once.
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     */
    public void render(GL gl, GLU glu)
    {
        // Ignored
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     */
    public void postRender(GL gl, GLU glu)
    {
        // Ignored
    }
}
