/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes a a fog rendering effect.
 * <p>
 *
 * All types of fog are reprensented in a single class. If the parameter
 * type is not useful for the fog mode, it is ignored.
 *
 * @author Justin Couch
 * @version $Revision: 1.9 $
 */
public class Fog extends Leaf
{
    /** Message when the linear ranges are reversed */
    private static final String FOG_RANGE_MSG =
        "The start distance is greater than the end distance";

    /** Set the fog mode to exponential curve */
    public static final int EXPONENTIAL = GL.GL_EXP;

    /** Set the fog mode to exponential-squared curve */
    public static final int EXPONENTIAL_2 = GL.GL_EXP2;

    /** Set the fog mode to linear */
    public static final int LINEAR = GL.GL_LINEAR;

    /** Has the texture attribute object reference changed */
    private boolean fogChanged;

    /** The current mode in use */
    private int mode;

    /** Density function if this is an exponential type */
    private float density;

    /** Start distance parameter for linear fog */
    private float startDistance;

    /** End (max density) distance parameter for linear fog */
    private float endDistance;

    /** Base colour of the fog */
    private float[] color;

    /**
     * Constructs a fog node with the default mode set to linear and the
     * colour set to white.
     */
    public Fog()
    {
        this(LINEAR);
    }

    /**
     * Construct a fog using the given mode and the colour set to white.
     */
    public Fog(int mode)
    {
        this(mode, null);
    }

    /**
     * Construct a fog using the given mode and colour.
     */
    public Fog(int mode, float[] c)
    {
        this.mode = mode;
        endDistance = 1;

        color = new float[3];

        if(c == null)
        {
            color[0] = 1;
            color[1] = 1;
            color[2] = 1;
        }
        else
        {
            color[0] = c[0];
            color[1] = c[1];
            color[2] = c[2];
        }
    }

    //----------------------------------------------------------
    // Methods overriding the Node class
    //----------------------------------------------------------

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void render(GL gl, GLU glu)
    {
        gl.glFogi(GL.GL_FOG_MODE, mode);
        gl.glFogfv(GL.GL_FOG_COLOR, color);

        switch(mode)
        {
            case LINEAR:
                gl.glFogf(GL.GL_FOG_START, startDistance);
                gl.glFogf(GL.GL_FOG_END, endDistance);
                break;

            case EXPONENTIAL:
                gl.glFogf(GL.GL_FOG_DENSITY, density);
                break;

            case EXPONENTIAL_2:
                gl.glFogf(GL.GL_FOG_DENSITY, density);
                break;
        }

        gl.glEnable(GL.GL_FOG);
    }

    /**
     * Restore all openGL state to the given drawable.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(GL gl, GLU glu)
    {
        gl.glDisable(GL.GL_FOG);
    }

    /**
     * Mark this node as having dirty bounds due to one of it's children having
     * their bounds changed.
     */
    protected void markBoundsDirty()
    {
        // Bounds are not used by Fog.
    }

    //----------------------------------------------------------
    // Local methods
    //----------------------------------------------------------

    /**
     * Change the fog colour to the new colour. Fog colour only takes RGB.
     *
     * @param c The colour to copy in
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setColor(float[] c)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        color[0] = c[0];
        color[1] = c[1];
        color[2] = c[2];
    }

    /**
     * Get the current drawing colour
     *
     * @param c An array of length 3 or more to copy the colour to
     */
    public void getColor(float[] c)
    {
        c[0] = color[0];
        c[1] = color[1];
        c[2] = color[2];
    }

    /**
     * Set the distance functions for the fog when in linear mode.
     *
     * @param start The closest distance that fog starts at
     * @param end The distance that the fog is fully opaque at
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLinearDistance(float start, float end)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(start > end)
            throw new IllegalArgumentException(FOG_RANGE_MSG);

        startDistance = start;
        endDistance = end;
    }

    /**
     * Get the two distance values for the linear fog settings.
     * The values are returned in the provided array with index 0 being the
     * start and index 1 being the end.
     *
     * @param values An array to copy the values into
     */
    public void getLinearDistance(float[] values)
    {
        values[0] = startDistance;
        values[1] = endDistance;
    }

    /**
     * Set the exponential decay factor.
     *
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setDensityRate(float rate)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        density = rate;
    }

    /**
     * Get the current decay rate.
     *
     * @return a value > 0
     */
    public float getDensityRate()
    {
        return density;
    }
}