/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the material properties of an object.
 *
 * @author Alan Hudson
 * @version $Revision: 1.17 $
 */
public class Material extends NodeComponent
    implements Comparable
{
    /** Default diffuse colour */
    private static final float[] DEFAULT_DIFFUSE = {0.8f, 0.8f, 0.8f};

    /** Default ambient colour */
    private static final float[] DEFAULT_AMBIENT = {0.2f, 0.2f, 0.2f};

    /** Default emissive colour */
    private static final float[] DEFAULT_EMISSIVE = {0, 0, 0};

    /** Default specular colour */
    private static final float[] DEFAULT_SPECULAR = {0, 0, 0};

    /** Default shininess colour */
    private static final float DEFAULT_SHININESS = 0.2f;

    /** Default transparency value */
    private static final float DEFAULT_TRANSPARENCY = 1;

    /** Should lighting be used for this material? */
    private boolean useLighting;

    /** Diffuse colour components */
    private float[] diffuseColor;

    /** Ambient colour components */
    private float[] ambientColor;

    /** Emissive colour components */
    private float[] emissiveColor;

    /** Specular colour components */
    private float[] specularColor;

    /** shinine colour factor */
    private float shininess;

    /** A mapping between glContext and displayListID(Integer) */
    private HashMap displayListMap;

    /** Flag to say the state has changed in the material */
    private boolean stateChanged;

    /**
     * Constructs a material with default values.
     * <pre>
     * ambient color = (0.2, 0.2, 0.2)
     * emisive color = (0.0, 0.0, 0.0)
     * diffuse color = (0.8, 0.8, 0.8)
     * specular color = (0, 0, 0)
     * shininess = 0.2
     * transparency = 1
     * </pre>
     */
    public Material()
    {
        this(DEFAULT_AMBIENT,
             DEFAULT_EMISSIVE,
             DEFAULT_DIFFUSE,
             DEFAULT_SPECULAR,
             DEFAULT_SHININESS,
             DEFAULT_TRANSPARENCY);
    }

    /**
     * Create a new material with all the colours specified.
     *
     * @param ambientColor
     * @param emissiveColor
     * @param diffuseColor
     * @param specularColor
     * @param shininess The shininess factor value
     * @param transparency A transparency value between 0 and 1
     */
    public Material(float[] ambientColor,
                    float[] emissiveColor,
                    float[] diffuseColor,
                    float[] specularColor,
                    float shininess,
                    float transparency)
    {
        this.diffuseColor = new float[4];
        this.diffuseColor[0] = diffuseColor[0];
        this.diffuseColor[1] = diffuseColor[1];
        this.diffuseColor[2] = diffuseColor[2];
        this.diffuseColor[3] = transparency;

        this.ambientColor = new float[3];
        this.ambientColor[0] = ambientColor[0];
        this.ambientColor[1] = ambientColor[1];
        this.ambientColor[2] = ambientColor[2];

        this.specularColor = new float[3];
        this.specularColor[0] = specularColor[0];
        this.specularColor[1] = specularColor[1];
        this.specularColor[2] = specularColor[2];

        this.emissiveColor = new float[3];
        this.emissiveColor[0] = emissiveColor[0];
        this.emissiveColor[1] = emissiveColor[1];
        this.emissiveColor[2] = emissiveColor[2];

        // Convert [0,1] to [0,128] for openGL
        this.shininess = 128 * shininess;
        useLighting = true;

        displayListMap = new HashMap(1);
        stateChanged = true;

    }

    //---------------------------------------------------------------
    // Methods defined by NodeComponent
    //---------------------------------------------------------------

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // If we have changed state, then clear the old display lists
        if(stateChanged)
        {
            synchronized(displayListMap)
            {
                displayListMap.clear();
                stateChanged = false;
            }
        }

        Integer listName = (Integer)displayListMap.get(gl);

        if(listName == null)
        {
            listName = new Integer(gl.glGenLists(1));

            gl.glNewList(listName.intValue(), GL.GL_COMPILE_AND_EXECUTE);

            if(useLighting)
                gl.glEnable(GL.GL_LIGHTING);

            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, diffuseColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, ambientColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, specularColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_EMISSION, emissiveColor);
            gl.glMaterialf(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, shininess);

            gl.glEndList();
            displayListMap.put(gl, listName);
        }
        else
        {
            gl.glCallList(listName.intValue());
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        if(useLighting)
            gl.glDisable(GL.GL_LIGHTING);
    }

    //---------------------------------------------------------------
    // Methods defined by Comparable
    //---------------------------------------------------------------

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     * <p>
     *
     * Derived classes should extend this one to add the extra comparisons
     * needed.
     *
     * @param o The object to be compared
     * @return -1, 0 or 1 depending on order
     * @throws ClassCastException The specified object's type prevents it from
     *    being compared to this Object
     */
    public int compareTo(Object o)
        throws ClassCastException
    {
        Material mat = (Material)o;
        return compareTo(mat);
    }

    //---------------------------------------------------------------
    // Methods defined by Object
    //---------------------------------------------------------------

    /**
     * Compare this object for equality to the given object.
     *
     * @param o The object to be compared
     * @return True if these represent the same values
     */
    public boolean equals(Object o)
    {
        if(!(o instanceof Material))
            return false;
        else
            return equals((Material)o);

    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Set the shininess factor. The shininess should be a vlue between 0 and
     * 1.
     *
     * @param s The shininess factor value
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setShininess(float s)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        // Convert [0,1] to [0,128] for openGL
        shininess = 128 * s;
        stateChanged = true;
    }

    /**
     * Get the current shininess value.
     *
     * @return a number between 0 and 1
     */
    public float getShininess()
    {
        return shininess;
    }

    /**
     * Set the diffuse color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setDiffuseColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        diffuseColor[0] = col[0];
        diffuseColor[1] = col[1];
        diffuseColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Get the current value of the diffuse color.
     *
     * @param col The array to copy the values into
     */
    public void getDiffuseColor(float[] col)
    {
        col[0] = diffuseColor[0];
        col[1] = diffuseColor[1];
        col[2] = diffuseColor[2];
    }

    /**
     * Set the specular color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setSpecularColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        specularColor[0] = col[0];
        specularColor[1] = col[1];
        specularColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Get the current value of the specular color.
     *
     * @param col The array to copy the values into
     */
    public void getSpecularColor(float[] col)
    {
        col[0] = specularColor[0];
        col[1] = specularColor[1];
        col[2] = specularColor[2];
    }

    /**
     * Set the emissive color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setEmissiveColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        emissiveColor[0] = col[0];
        emissiveColor[1] = col[1];
        emissiveColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Get the current value of the emissive color.
     *
     * @param col The array to copy the values into
     */
    public void getEmissiveColor(float[] col)
    {
        col[0] = emissiveColor[0];
        col[1] = emissiveColor[1];
        col[2] = emissiveColor[2];
    }

    /**
     * Set the ambient color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAmbientColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        ambientColor[0] = col[0];
        ambientColor[1] = col[1];
        ambientColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Get the current value of the ambient color.
     *
     * @param col The array to copy the values into
     */
    public void getAmbientColor(float[] col)
    {
        col[0] = ambientColor[0];
        col[1] = ambientColor[1];
        col[2] = ambientColor[2];
    }

    /**
     * Set the amount of transparency to be used for this material.
     *
     * @param transparency A value between 0 and 1
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTransparency(float transparency)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        diffuseColor[3] = transparency;
        stateChanged = true;
    }

    /**
     * Get the value for the existing transparency values
     *
     * @return The current transparency setting
     */
    public float getTransparency()
    {
        return diffuseColor[3];
    }

    /**
     * Set the flag that enables or disables lighting on any geometry using
     * this material.
     *
     * @param state true if lighting should be used
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLightingEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(useLighting != state)
        {
            useLighting = state;
            stateChanged = true;
        }
    }

    /**
     * Get the current lighting enabled flag.
     */
    public boolean isLightingEnabled()
    {
        return useLighting;
    }

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     *
     * @param mat The material to be compared
     * @return -1, 0 or 1 depending on order
     */
    public int compareTo(Material mat)
    {
        if(mat == this)
            return 0;

        if(mat == null)
            return 1;

        // check the material properties. Transparency is ignored
        // because we assume that the check before this has weeded
        // out the transparent from non-transparent for separate
        // sorting policies.
        if(useLighting != mat.useLighting)
            return !useLighting ? -1 : 1;

        if(shininess != mat.shininess)
            return shininess > mat.shininess ? 1 : -1;

        int res = compareColor3(ambientColor, mat.ambientColor);
        if(res != 0)
            return res;

        res = compareColor3(diffuseColor, mat.diffuseColor);
        if(res != 0)
            return res;

        res = compareColor3(emissiveColor, mat.emissiveColor);
        if(res != 0)
            return res;

        res = compareColor3(specularColor, mat.specularColor);
        if(res != 0)
            return res;

        return 0;
    }

    /**
     * Compares this object with the specified object to check for equivalence.
     *
     * @param mat The material to be compared
     * @return true if the objects represent identical values
     */
    public boolean equals(Material mat)
    {
        if(mat == this)
            return true;

        if((mat == null) ||
           (useLighting != mat.useLighting) ||
           (shininess != mat.shininess) ||
           !equalsColor3(ambientColor, mat.ambientColor) ||
           !equalsColor3(diffuseColor, mat.diffuseColor) ||
           !equalsColor3(emissiveColor, mat.emissiveColor) ||
           !equalsColor3(specularColor, mat.specularColor))
            return false;

        return true;
    }

    /**
     * Compare 2 color arrays of length 3 for equality
     *
     * @param a The first colour array to check
     * @param b The first colour array to check
     * @return -1 if a[i] < b[i], +1 if a[i] > b[i], otherwise 0
     */
    private int compareColor3(float[] a, float[] b)
    {
        if(a[0] < b[0])
            return -1;
        else if (a[0] > b[0])
            return 1;

        if(a[1] < b[1])
            return -1;
        else if (a[1] > b[1])
            return 1;

        if(a[2] < b[2])
            return -1;
        else if (a[2] > b[2])
            return 1;

        return 0;
    }

    /**
     * Compare 2 color arrays of length 3 for equality
     *
     * @param a The first colour array to check
     * @param b The first colour array to check
     * @return true if they have the same values, false otherwise
     */
    private boolean equalsColor3(float[] a, float[] b)
    {
        return (a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]);
    }
}