/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the basic textured appearance of an object.
 * <p>
 *
 * This is the base class for all texture objects used in Aviatrix3D. It does
 * not provide any functional capabilities, just a collection of the common
 * constants and state. It provides abilities based on the minimal setup for
 * a 1-dimensional texture. For settings for addtional dimensions (T and R)
 * please visit the appropriate derived class.
 * <p>
 * All textures default to the following setup during the constructor call:
 * <ul>
 * <li>Boundary Mode: Clamped</li>
 * <li>Min Filter: Fastest</li>
 * <li>Mag Filter: Fastest</li>
 * <li>Anisotropic Mode: None</li>
 * <li>Mipmap Mode: None (Base level only)</li>
 * </ul>
 *
 * @author Alan Hudson
 * @version $Revision: 1.21 $
 */
public abstract class Texture extends NodeComponent
    implements DeletableSceneGraphObject, Comparable
{
    /** MipMapMode constants - No Mip Map*/
    public static final int MODE_BASE_LEVEL = 0;

    /** MipMapMode constants - Use Mip Maps */
    public static final int MODE_MIPMAP = 1;

    // Boundary Modes
    /** Boundary mode to repeat textures */
    public static final int BM_WRAP = GL.GL_REPEAT;

    /** Boundary mode to clamp textures */
    public static final int BM_CLAMP = GL.GL_CLAMP;

    /** Boundary mode to clamp the texture edge value without border */
    public static final int BM_CLAMP_TO_EDGE = GL.GL_CLAMP_TO_EDGE;

    /** Boundary mode to clamp the texture border colour */
    public static final int BM_CLAMP_TO_BOUNDARY = GL.GL_CLAMP_TO_BORDER;
    //public static final int BM_MIRRORED_REPEAT = GL.GL_MIRRORED_REPEAT;

    // Maxification Filter Techniques

    /** Set the maginification filter to the fastest option */
    public static final int MAGFILTER_FASTEST = 0;

    /** Set the maginification filter to the highest quality option */
    public static final int MAGFILTER_NICEST = 1;

    /** Set the maginification filter to base point-base filtering */
    public static final int MAGFILTER_BASE_LEVEL_POINT = 2;

    /** Set the maginification filter to linear filtering */
    public static final int MAGFILTER_BASE_LEVEL_LINEAR = 3;

    /** Set the maginfication filter to use the detail texture option */
    public static final int MAGFILTER_LINEAR_DETAIL = 4;

    /** Set the maginfication filter to use the detail texture's RGB values */
    public static final int MAGFILTER_LINEAR_DETAIL_RGB = 5;

    /** Set the maginfication filter to use the detail texture's alpha values */
    public static final int MAGFILTER_LINEAR_DETAIL_ALPHA = 6;

    // Minification Filter Techniques
    /** Set the minification filter to the fastest option */
    public static final int MINFILTER_FASTEST = 0;

    /** Set the mininification filter to the highest quality  option */
    public static final int MINFILTER_NICEST = 1;

    /** Set the mininification filter to base point-base filtering */
    public static final int MINFILTER_BASE_LEVEL_POINT = 2;

    /** Set the mininification filter to linear filtering */
    public static final int MINFILTER_BASE_LEVEL_LINEAR = 3;

    /** Set the mininification filter to base point-base filtering */
    public static final int MINFILTER_MULTI_LEVEL_POINT = 4;

    /** Set the mininification filter to linear filtering */
    public static final int MINFILTER_MULTI_LEVEL_LINEAR = 5;

    // Anistropic Modes

    /** Disable anisotropic filtering */
    public static final int ANISOTROPIC_MODE_NONE = 0;

    /** Enable anisotropic filtering */
    public static final int ANISOTROPIC_MODE_SINGLE = 1;

    // Format values
    // More need to be added here later.

    /** Interpret the texture format as alpha only */
    public static final int FORMAT_ALPHA = GL.GL_ALPHA;

    /** Interpret the texture format as intensity only */
    public static final int FORMAT_INTENSITY = GL.GL_LUMINANCE;

    /** Interpret the texture format as intensity-alpha */
    public static final int FORMAT_INTENSITY_ALPHA = GL.GL_LUMINANCE_ALPHA;

    /** Interpret the texture format as RGB */
    public static final int FORMAT_RGB = GL.GL_RGB;

    /** Interpret the texture format as RGBA */
    public static final int FORMAT_RGBA = GL.GL_RGBA;

    /** The images defining this texture */
    protected TextureComponent[] images;

    /** The number of valid items in the image array */
    protected int numImages;

    /** The Anisotropic Filtering Mode */
    protected int anisotropicMode;

    /** The Anisotropic Filtering Degree */
    protected float anisotropicDegree;

    /** The magnification filter */
    protected int magFilter;

    /** The minification filter */
    protected int minFilter;

    /** The boundary mode S value */
    protected int boundaryModeS;

    /** The mipMapMode */
    protected int mipMapMode;

    /** The width of the main texture. */
    protected int width;

    /** The pixel format of the main texture image. */
    protected int format;

    /**
     * Flag to say that the display lists must be cleared and regenerated
     * because some state changed
     */
    protected boolean stateChanged;

    /** A mapping between glContext and displayListID(Integer) */
    protected HashMap displayListMap;

    /** The GL type of the texture. */
    protected final int textureType;

    /** The mapping of GL context to OpenGL texture ID*/
    protected HashMap textureIdMap;

    /**
     * Constructs a texture with default values. The mipmap mode is set to
     * MODE_BASE_LEVEL.
     *
     * @param type One of the texture type constants
     */
    protected Texture(int type)
    {
        width = -1;
        textureType = type;
        mipMapMode = MODE_BASE_LEVEL;
        boundaryModeS = BM_CLAMP;

        magFilter = MAGFILTER_FASTEST;
        minFilter = MINFILTER_FASTEST;

        textureIdMap = new HashMap();
        displayListMap = new HashMap(1);
    }

    //---------------------------------------------------------------
    // Methods defined by SceneGraphObject
    //---------------------------------------------------------------

    /**
     * Notification that this object is live now.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        if(state)
            liveCount++;
        else if(liveCount > 0)
            liveCount--;

        if((liveCount == 0) || !alive)
        {
            super.setLive(state);

            if(!state && updateHandler != null)
                updateHandler.objectRemoved(this);
        }
    }

    //---------------------------------------------------------------
    // Methods defined by DeletableSceneGraphObject
    //---------------------------------------------------------------

    /**
     * Cleanup the object now for the given GL context.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void cleanup(GL gl, GLU glu)
    {
        Integer t_id = (Integer)textureIdMap.get(gl);
        if(t_id != null)
        {
            int tex_id_tmp[] = { t_id.intValue() };
            gl.glDeleteTextures(1, tex_id_tmp);
            textureIdMap.remove(gl);
        }
    }

    //---------------------------------------------------------------
    // Methods defined by Comparable
    //---------------------------------------------------------------

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object. Derived instances
     * should override this to add texture-specific extensions.
     *
     * @param o The object to be compared
     * @return -1, 0 or 1 depending on order
     * @throws ClassCastException The specified object's type prevents it from
     *    being compared to this Object
     */
    public int compareTo(Object o)
        throws ClassCastException
    {
        Texture tex = (Texture)o;
        return compareTo(tex);
    }

    //---------------------------------------------------------------
    // Methods defined by Object
    //---------------------------------------------------------------

    /**
     * Compare this object for equality to the given object.
     *
     * @param o The object to be compared
     * @return True if these represent the same values
     */
    public boolean equals(Object o)
    {
        if(!(o instanceof Texture))
            return false;
        else
            return equals((Texture)o);

    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Set the images for this texture
     *
     * @param mipMapMode Flag stating the type of texture mode to use
     * @param format Image format to use for grayscale images
     * @param srcImages The image data to use, single for base level
     * @param num The valid number of images to use from the array
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setImages(int mipMapMode,
                          int format,
                          TextureComponent[] srcImages,
                          int num)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        // TODO: validity checking on these values
        width = num == 0 ? -1 : srcImages[0].getWidth();
        this.format = format;
        this.mipMapMode = mipMapMode;

        if(images == null || images.length < num)
            images = new TextureComponent[num];
        else
        {
            for(int i = numImages; i < num; i++)
                images[i] = null;
        }

        System.arraycopy(srcImages, 0, images, 0, num);
        numImages = num;
        stateChanged = true;
    }

    /**
     * Get the texture type. This returns one of the GL constant types that
     * represent the texture type - 1D, 2D, etc.
     *
     * @return The type constant of this texture
     */
    public int getTextureType()
    {
        return textureType;
    }

    /**
     * Get the format for this texture.
     *
     * @return The format.
     */
    public int getFormat()
    {
        return format;
    }

    /**
     * Get the width of the texture in pixels. If no image is set, this returns
     * -1.
     *
     * @return a number >= -1
     */
    public int getWidth()
    {
        return width;
    }

    /**
     * Set the aniostropic filtering mode.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAnisotropicFilterMode(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        anisotropicMode = mode;
        stateChanged = true;
    }

    /**
     * Get the current aniostropic filtering mode.
     *
     * @return The current mode.
     */
    public int getAnisotropicFilterMode()
    {
        return anisotropicMode;
    }

    /**
     * Set the anisotropic filtering degree.  Values greater
     * then the hardware supports will be clamped.
     *
     * @param degree The filtering degree.  1.0 is the default.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAnisotropicFilterDegree(float degree)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        anisotropicDegree = degree;

        // TODO: Need to clamp to hardware max
        stateChanged = true;
    }

    /**
     * Get the current anisotropic filtering degree. The value returned will
     * be the clamped maximum for the hardware support.
     *
     * @return The filtering degree.  1.0 is the default.
     */
    public float getAnisotropicFilterDegree()
    {
        return anisotropicDegree;
    }

    /**
     * Set the magnification filtering mode.
     *
     * @param mode The new mode.
     */
    public void setMagFilter(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        magFilter = mode;
        stateChanged = true;
    }

    /**
     * Get the magnification filtering mode.
     *
     * @return The current mode.
     */
    public int getMagFilter()
    {
        return magFilter;
    }

    /**
     * Set the magnification filtering mode.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setMinFilter(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        minFilter = mode;
        stateChanged = true;
    }

    /**
     * Get the minification filtering mode.
     *
     * @return The current mode.
     */
    public int getMinFilter()
    {
        return minFilter;
    }

    /**
     * Set the boundary handling for the S parameter.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBoundaryModeS(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        boundaryModeS = mode;
        stateChanged = true;
    }

    /**
     * Get the current boundary handling for the S parameter.
     *
     * @return The current mode.
     */
    public int getBoundaryModeS()
    {
        return boundaryModeS;
    }

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object. Derived instances
     * should override this to add texture-specific extensions.
     *
     * @param tex The texture instance to be compared
     * @return -1, 0 or 1 depending on order
     */
    public int compareTo(Texture tex)
    {
        if(tex == null)
            return 1;

        if(tex == this)
            return 0;

        if(textureType != tex.textureType)
            return textureType < tex.textureType ? -1 : 1;

        if(minFilter != tex.minFilter)
            return minFilter < tex.minFilter ? -1 : 1;

        if(magFilter != tex.magFilter)
            return magFilter < tex.magFilter ? -1 : 1;

        if(mipMapMode != tex.mipMapMode)
            return mipMapMode < tex.mipMapMode ? -1 : 1;

        if(anisotropicMode != tex.anisotropicMode)
            return anisotropicMode < tex.anisotropicMode ? -1 : 1;

        if(anisotropicDegree != tex.anisotropicDegree)
            return anisotropicDegree < tex.anisotropicDegree ? -1 : 1;

        if(format != tex.format)
            return format < tex.format ? -1 : 1;

        if(width != tex.width)
            return width < tex.width ? -1 : 1;

        if(numImages != tex.numImages)
            return numImages < tex.numImages ? -1 : 1;
        else
        {
            for(int i = 0; i < numImages; i++)
            {
                if(images[i] != tex.images[i])
                    return images[i].hashCode() < tex.images[i].hashCode() ? -1 : 1;
            }
        }

        return 0;
    }

    /**
     * Compares this object with the specified object to check for equivalence.
     *
     * @param tex The texture instance to be compared
     * @return true if the objects represent identical values
     */
    public boolean equals(Texture tex)
    {
        if(tex == this)
            return true;

        if((tex == null) ||
           (textureType != tex.textureType) ||
           (minFilter != tex.minFilter) ||
           (magFilter != tex.magFilter) ||
           (mipMapMode != tex.mipMapMode) ||
           (anisotropicMode != tex.anisotropicMode) ||
           (anisotropicDegree != tex.anisotropicDegree) ||
           (format != tex.format) ||
           (width != tex.width) ||
           (numImages != tex.numImages))
            return false;

        for(int i = 0; i < numImages; i++)
        {
            if(images[i] != tex.images[i])
                return false;
        }

        return true;
    }
}