/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.nio.ByteBuffer;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the 1D texture properties of an object.
 *
 * @author Justin Couch
 * @version $Revision: 1.4 $
 */
public class Texture1D extends Texture
{
    /**
     * Constructs a texture with default values.
     */
    public Texture1D()
    {
        super(GL.GL_TEXTURE_1D);
    }

    /**
     * Constructs a texture using just a single image, thus setting it initially
     * as MODE_BASE_LEVEL.
     */
    public Texture1D(int format, TextureComponent1D singleImage)
    {
        super(GL.GL_TEXTURE_1D);

        this.format = format;
        images = new TextureComponent[1];
        images[0] = singleImage;

        width = singleImage.getWidth();
        stateChanged = true;
        numImages = 1;
    }

    /**
     * Constructs a 1D texture with the specified images.  The number of images
     * should be 1 for no mip maps, or all mipmap levels should be provided.
     *
     * @param srcImages The image defining this texture.
     * @param num The number of valid images to take from the array
     */
    public Texture1D(int format, TextureComponent1D[] srcImages, int num)
    {
        super(GL.GL_TEXTURE_1D);

        this.format = format;
        if(srcImages != null)
        {
            if(num == 1)
                mipMapMode = MODE_BASE_LEVEL;
            else
                mipMapMode = MODE_MIPMAP;
        }

        images = new TextureComponent[num];
        System.arraycopy(srcImages, 0, images, 0, num);

        width = images[0].getWidth();
        stateChanged = true;
        numImages = num;
    }

    //---------------------------------------------------------------
    // Methods defined by NodeComponent
    //---------------------------------------------------------------

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(numImages == 0)
            return;

        if(stateChanged)
        {
            synchronized(textureIdMap)
            {
                textureIdMap.clear();
                stateChanged = false;
            }
        }

        Integer t_id = (Integer)textureIdMap.get(gl);
        if(t_id != null)
        {
            gl.glBindTexture(textureType, t_id.intValue());
            return;
        }

        // NOTE:
        // Assume, for now, that we need to regenerate the entire texture
        // because the imageComponent has changed. Once sub-image update is
        // implemented, we may want to change this logic
        int[] tex_id_tmp = new int[1];
        gl.glGenTextures(1, tex_id_tmp);
        textureIdMap.put(gl, new Integer(tex_id_tmp[0]));

        gl.glBindTexture(textureType, tex_id_tmp[0]);

        gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1);
        gl.glTexParameteri(GL.GL_TEXTURE_1D,
                           GL.GL_TEXTURE_WRAP_S,
                           boundaryModeS);

        int mode = 0;
        switch(magFilter)
        {
            case MAGFILTER_FASTEST:
            case MAGFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MAGFILTER_NICEST:
            case MAGFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            default:
                System.out.println("Unknown mode in MagFilter: " + magFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_1D,
                           GL.GL_TEXTURE_MAG_FILTER,
                           mode);

        switch(minFilter)
        {
            case MINFILTER_FASTEST:
            case MINFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MINFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_LINEAR:
                mode = GL.GL_LINEAR_MIPMAP_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_POINT:
                mode = GL.GL_NEAREST_MIPMAP_NEAREST;
                break;

            case MINFILTER_NICEST:
                mode = (numImages > 1) ? GL.GL_LINEAR_MIPMAP_LINEAR : GL.GL_LINEAR;
                break;

            default:
                System.out.println("Unknown mode in MinFilter: " + minFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_1D,
                           GL.GL_TEXTURE_MIN_FILTER,
                           mode);

        if(anisotropicMode != ANISOTROPIC_MODE_NONE)
        {
            // float[] val = new float[1];

            //gl.glGetFloatv(GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, val);
            gl.glTexParameterf(GL.GL_TEXTURE_1D,
                               GL.GL_TEXTURE_MAX_ANISOTROPY_EXT,
                               anisotropicDegree);
        }

        int img_count = (mipMapMode == MODE_BASE_LEVEL) ? 1 : numImages;

        for(int i = 0; i < img_count; i++)
        {
            ByteBuffer pixels = images[i].getData(i);

            int width = images[i].getWidth();
            int comp_format = images[i].getFormat(i);
            int int_format = GL.GL_RGB;
            int ext_format = GL.GL_RGB;

            switch(comp_format)
            {
                case TextureComponent.FORMAT_RGB:
                    int_format = GL.GL_RGB;
                    ext_format = GL.GL_RGB;
                    break;

                case TextureComponent.FORMAT_RGBA:
                    int_format = GL.GL_RGBA;
                    ext_format = GL.GL_RGBA;
                    break;

                case TextureComponent.FORMAT_BGR:
                    int_format = GL.GL_BGR;
                    ext_format = GL.GL_BGR;
                    break;

                case TextureComponent.FORMAT_BGRA:
                    int_format = GL.GL_BGRA;
                    ext_format = GL.GL_BGRA;
                    break;


                case TextureComponent.FORMAT_INTENSITY_ALPHA:
                    int_format = GL.GL_LUMINANCE_ALPHA;
                    ext_format = GL.GL_LUMINANCE_ALPHA;
                    break;

                case TextureComponent.FORMAT_SINGLE_COMPONENT:
                    switch(format)
                    {
                        case FORMAT_INTENSITY:
                            int_format = GL.GL_LUMINANCE;
                            ext_format = GL.GL_LUMINANCE;
                            break;

                        case FORMAT_ALPHA:
                            int_format = GL.GL_ALPHA;
                            ext_format = GL.GL_ALPHA;
                    }
                    break;

                default:
            }

            gl.glTexImage1D(GL.GL_TEXTURE_1D,
                            i,
                            int_format,
                            width,
                            0,
                            ext_format,
                            GL.GL_UNSIGNED_BYTE,
                            pixels);
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        gl.glDisable(textureType);
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object. Derived instances
     * should override this to add texture-specific extensions.
     *
     * @param tex The texture instance to be compared
     * @return -1, 0 or 1 depending on order
     */
    public int compareTo(Texture tex)
    {
        int res = super.compareTo(tex);
        if(res != 0)
            return res;

        Texture1D t1d = (Texture1D)tex;

        if(boundaryModeS != t1d.boundaryModeS)
            return boundaryModeS < t1d.boundaryModeS ? -1 : 1;

        return 0;
    }

    /**
     * Compares this object with the specified object to check for equivalence.
     *
     * @param tex The texture instance to be compared
     * @return true if the objects represent identical values
     */
    public boolean equals(Texture tex)
    {
        if(!super.equals(tex))
            return false;

        if(!(tex instanceof Texture1D))
            return false;

        Texture1D t1d = (Texture1D)tex;

        return (boundaryModeS == t1d.boundaryModeS);
    }
}