/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.nio.ByteBuffer;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * A specialist object that renders a cubic environment map from pre-built
 * images.
 * <p>
 *
 * All textures must be square in size. If they are not square, then an error
 * is generated.
 * <p>
 *
 * This implementation does not handle dynamic cubic environment mapping, nor
 * mipmaps.
 *
 * @author Justin Couch
 * @version $Revision: 1.15 $
 */
public class TextureCubicEnvironmentMap extends Texture
{
    /** The texture belongs to the positive X axis */
    public static final int POSITIVE_X = 0;

    /** The texture belongs to the negative X axis */
    public static final int NEGATIVE_X = 1;

    /** The texture belongs to the positive Y axis */
    public static final int POSITIVE_Y = 2;

    /** The texture belongs to the negative Y axis */
    public static final int NEGATIVE_Y = 3;

    /** The texture belongs to the positive Z axis */
    public static final int POSITIVE_Z = 4;

    /** The texture belongs to the negative Z axis */
    public static final int NEGATIVE_Z = 5;

    /** Internal array indexing these constants to GL constants */
    private static final int[] TEXTURE_TARGETS =
    {
        GL.GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
        GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
        GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
        GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
        GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
        GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
    };

    /** The height of the main texture. */
    protected int height;

    /** The boundary mode S value */
    protected int boundaryModeT;

    /**
     * Constructs a texture with default values.
     */
    public TextureCubicEnvironmentMap()
    {
        super(GL.GL_TEXTURE_CUBE_MAP_ARB);

        images = new TextureComponent[6];
        height = -1;

        boundaryModeT = BM_CLAMP;
    }

    /**
     * Constructs a texture with the given list of images. The provided array
     * must be at least 6 items long, even if some are null. The order taken
     * for the images must follow the constants defined for the image IDs.
     * MipMaps are not handled in this case.
     */
    public TextureCubicEnvironmentMap(TextureComponent2D[] srcImages)
    {
        this();

        // TODO:
        // Need to set this properly and make sure that the images are
        // square.
        height = -1;

        if(srcImages != null)
        {
            this.images[0] = srcImages[0];
            this.images[1] = srcImages[1];
            this.images[2] = srcImages[2];
            this.images[3] = srcImages[3];
            this.images[4] = srcImages[4];
            this.images[5] = srcImages[5];

            stateChanged = true;
        }
    }

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(numImages == 0)
            return;

        if(stateChanged)
        {
            synchronized(textureIdMap)
            {
                textureIdMap.clear();
                stateChanged = false;
            }
        }

        Integer t_id = (Integer)textureIdMap.get(gl);
        if(t_id != null)
        {
            gl.glBindTexture(textureType, t_id.intValue());
            return;
        }

        // NOTE:
        // Assume, for now, that we need to regenerate the entire texture
        // because the imageComponent has changed. Once sub-image update is
        // implemented, we may want to change this logic
        int[] tex_id_tmp = new int[1];
        gl.glGenTextures(1, tex_id_tmp);
        textureIdMap.put(gl, new Integer(tex_id_tmp[0]));

        gl.glBindTexture(textureType, tex_id_tmp[0]);

        // boundary modes are ignored on Cube maps
        int mode = 0;
        switch(magFilter)
        {
            case MAGFILTER_FASTEST:
            case MAGFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MAGFILTER_NICEST:
            case MAGFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            default:
                System.out.println("Unknown mode in MagFilter: " + magFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP_ARB,
                           GL.GL_TEXTURE_MAG_FILTER,
                           mode);

        switch(minFilter)
        {
            case MINFILTER_FASTEST:
            case MINFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MINFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_LINEAR:
                mode = GL.GL_LINEAR_MIPMAP_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_POINT:
                mode = GL.GL_NEAREST_MIPMAP_NEAREST;
                break;

            case MINFILTER_NICEST:
                mode = (numImages > 1) ? GL.GL_LINEAR_MIPMAP_LINEAR : GL.GL_LINEAR;
                break;

            default: System.out.println("Unknown mode in MinFilter: " + minFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP_ARB,
                           GL.GL_TEXTURE_MIN_FILTER,
                           mode);

        if(anisotropicMode != ANISOTROPIC_MODE_NONE)
        {
            // float[] val = new float[1];

            //gl.glGetFloatv(GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, val);
            gl.glTexParameterf(GL.GL_TEXTURE_CUBE_MAP_ARB,
                               GL.GL_TEXTURE_MAX_ANISOTROPY_EXT,
                               anisotropicDegree);
        }

        gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1);

        // Does not support mipmapped levels yet.

        for(int i = 0; i < 6; i++)
        {
            if(images[i] == null)
                continue;

            ByteBuffer pixels = images[i].getData(0);

            int width = images[i].getWidth();
            int height = ((TextureComponent2D)images[i]).getHeight();
            int comp_format = images[i].getFormat(0);
            int int_format = GL.GL_RGB;
            int ext_format = GL.GL_RGB;

            switch(comp_format)
            {
                case TextureComponent.FORMAT_RGB:
                    int_format = GL.GL_RGB;
                    ext_format = GL.GL_RGB;
                    break;

                case TextureComponent.FORMAT_RGBA:
                    int_format = GL.GL_RGBA;
                    ext_format = GL.GL_RGBA;
                    break;

                case TextureComponent.FORMAT_BGR:
                    int_format = GL.GL_BGR;
                    ext_format = GL.GL_BGR;
                    break;

                case TextureComponent.FORMAT_BGRA:
                    int_format = GL.GL_BGRA;
                    ext_format = GL.GL_BGRA;
                    break;

                case TextureComponent.FORMAT_INTENSITY_ALPHA:
                    int_format = GL.GL_LUMINANCE_ALPHA;
                    ext_format = GL.GL_LUMINANCE_ALPHA;
                    break;

                case TextureComponent.FORMAT_SINGLE_COMPONENT:
                    switch(format)
                    {
                        case FORMAT_INTENSITY:
                            int_format = GL.GL_LUMINANCE;
                            ext_format = GL.GL_LUMINANCE;
                            break;

                        case FORMAT_ALPHA:
                            int_format = GL.GL_ALPHA;
                            ext_format = GL.GL_ALPHA;
                    }
                    break;
            }

            gl.glTexImage2D(TEXTURE_TARGETS[i],
                            0,
                            int_format,
                            width,
                            height,
                            0,
                            ext_format,
                            GL.GL_UNSIGNED_BYTE,
                            pixels);
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
         gl.glDisable(textureType);
    }

    /**
     * Set the boundary handling for the S parameter.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBoundaryModeT(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        boundaryModeT = mode;
        stateChanged = true;
    }

    /**
     * Get the current boundary handling for the S parameter.
     *
     * @return The current mode.
     */
    public int getBoundaryModeT()
    {
        return boundaryModeT;
    }

    /**
     * Get the height of the texture in pixels. If no image is set, this returns
     * -1.
     *
     * @return a number >= -1
     */
    public int getHeight()
    {
        return height;
    }

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object. Derived instances
     * should override this to add texture-specific extensions.
     *
     * @param tex The texture instance to be compared
     * @return -1, 0 or 1 depending on order
     */
    public int compareTo(Texture tex)
    {
        int res = super.compareTo(tex);
        if(res != 0)
            return res;

        TextureCubicEnvironmentMap cem = (TextureCubicEnvironmentMap)tex;

        if(height != cem.height)
            return height < cem.height ? -1 : 1;

        if(boundaryModeS != cem.boundaryModeS)
            return boundaryModeS < cem.boundaryModeS ? -1 : 1;

        if(boundaryModeT != cem.boundaryModeT)
            return boundaryModeT < cem.boundaryModeT ? -1 : 1;

        return 0;
    }

    /**
     * Compares this object with the specified object to check for equivalence.
     *
     * @param tex The texture instance to be compared
     * @return true if the objects represent identical values
     */
    public boolean equals(Texture tex)
    {
        if(!super.equals(tex))
            return false;

        if(!(tex instanceof Texture2D))
            return false;

        TextureCubicEnvironmentMap cem = (TextureCubicEnvironmentMap)tex;

        if((height != cem.height) ||
           (boundaryModeS != cem.boundaryModeS) ||
           (boundaryModeT != cem.boundaryModeT))
            return false;

        return true;
    }
}