/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * An OpenGL TriangleFanArray.
 *
 * @author Justin Couch
 * @version $Revision: 1.15 $
 */
public class TriangleFanArray extends VertexGeometry
{
    /** Message when the fan count is < 3 */
    private static final String MIN_COUNT_MSG =
        "A fan with less than 3 vertices is specified at index ";

    /** Message when the total count is less than fan totals */
    private static final String TOTAL_COUNT_MSG =
        "The total requested indexes in the fan is greater than the " +
        "number of valid vertices provided.";

    /** The fan counts of valid triangles for each fan */
    private int[] fanCounts;

    /** The number of valid fans to read from the array */
    private int numFans;

    /**
     * Constructs a TriangleFanArray with default values.
     */
    public TriangleFanArray()
    {
        initPolygonDetails(3);
    }

    /**
     * Set the number of valid fans to use. A check is performed to make
     * sure that the number of vertices high enough to support the total
     * of all the fan counts so make sure to call setVertex() with the
     * required array length before calling this method. Each fan must be
     * a minumum length of three.
     *
     * @param counts The array of counts
     * @param num The number of valid items to read from the array
     * @throws IllegalArgumentException Invalid total fan count or
     *   individual fan count < 3
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setFanCount(int[] counts, int num)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        int total = 0;

        for(int i = 0; i < num; i++)
        {
            if(counts[i] < 3)
                throw new IllegalArgumentException(MIN_COUNT_MSG + i);

            total += counts[i];
        }

        if(fanCounts == null || fanCounts.length < num)
            fanCounts = new int[num];

        if(total > numCoords)
            throw new IllegalArgumentException(TOTAL_COUNT_MSG);

        numFans = num;

        if(num > 0)
            System.arraycopy(counts, 0, fanCounts, 0, num);
    }

    /**
     * Get the number of valid fans that are defined for this geometry.
     *
     * @return a positive number
     */
    public int getValidFanCount()
    {
        return numFans;
    }

    /**
     * Get the sizes of the valid fans. The passed array must be big enough
     * to contain all the fan counts.
     *
     * @param values An array to copy the strip values into
     */
    public void getFanCount(int[] values)
    {
        System.arraycopy(fanCounts, 0, values, 0, numFans);
    }

    /**
     * Issue ogl commands needed for this renderable object.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // No coordinates, do nothing.
        if((numFans == 0) || ((vertexFormat & COORDINATE_MASK) == 0))
            return;

        //  State enable first
        int vtx_size = vertexFormat & 0x07;
        gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
        gl.glVertexPointer(vtx_size, GL.GL_FLOAT, 0, vertexBuffer);

        if((vertexFormat & NORMALS) != 0)
        {
            gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
            gl.glNormalPointer(GL.GL_FLOAT, 0, normalBuffer);
        }

        if((vertexFormat & TEXTURE_MASK) != 0)
        {
            // single texturing or multi-texturing
            if(numTextureSets == 1)
            {
                gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
                gl.glTexCoordPointer(textureTypes[0],
                                     GL.GL_FLOAT,
                                     0,
                                     textureBuffer[0]);
            }
            else
            {
                for(int i = 0; i < numTextureSets; i++)
                {
                    int set_id = textureSets[i];
                    gl.glClientActiveTexture(GL.GL_TEXTURE0 + i);
                    gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
                    gl.glTexCoordPointer(textureTypes[set_id],
                                         GL.GL_FLOAT,
                                         0,
                                         textureBuffer[set_id]);
                }
            }
        }

        if((vertexFormat & COLOR_MASK) != 0)
        {
            gl.glEnableClientState(GL.GL_COLOR_ARRAY);
            int size = ((vertexFormat & COLOR_3) != 0) ? 3 : 4;
            gl.glColorPointer(size, GL.GL_FLOAT, 0, colorBuffer);
        }

        if((vertexFormat & COLOR2_MASK) != 0)
        {
            gl.glEnableClientState(GL.GL_SECONDARY_COLOR_ARRAY);
            gl.glSecondaryColorPointer(3, GL.GL_FLOAT, 0, color2Buffer);
        }

        if((vertexFormat & FOG_MASK) != 0)
        {
            gl.glFogi(GL.GL_FOG_COORDINATE_SOURCE, GL.GL_FOG_COORDINATE);
            gl.glEnableClientState(GL.GL_FOG_COORDINATE_ARRAY);
            gl.glFogCoordPointer(GL.GL_FLOAT, 0, fogBuffer);
        }

        // Loop through the provided arrays using the sub-range
        // draw capabilities of glDrawArrays to create multiple
        // fans being drawn.
        int fan_offset = 0;
        for(int i = 0; i < numFans; i++)
        {
            gl.glDrawArrays(GL.GL_TRIANGLE_FAN,
                            fan_offset,
                            fanCounts[i]);

            fan_offset += fanCounts[i] * 3;
        }

        if((vertexFormat & FOG_MASK) != 0)
        {
            gl.glDisableClientState(GL.GL_FOG_COORDINATE_ARRAY);
            gl.glFogi(GL.GL_FOG_COORDINATE_SOURCE, GL.GL_FRAGMENT_DEPTH);
        }

        if((vertexFormat & COLOR2_MASK) != 0)
        {
            gl.glDisableClientState(GL.GL_SECONDARY_COLOR_ARRAY);
            gl.glSecondaryColor3f(1f,1f,1f);
        }

        if((vertexFormat & COLOR_MASK) != 0)
        {
            gl.glDisableClientState(GL.GL_COLOR_ARRAY);
            gl.glColor4f(1f,1f,1f,1f);
        }

        if((vertexFormat & TEXTURE_MASK) != 0)
        {
            // single texturing or multi-texturing
            if(numTextureSets == 1)
            {
                gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);
            }
            else
            {
                for(int i = numTextureSets - 1; i >= 0; i--)
                {
                    gl.glClientActiveTexture(GL.GL_TEXTURE0 + i);
                    gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);
                }
            }
        }

        if((vertexFormat & NORMALS) != 0)
            gl.glDisableClientState(GL.GL_NORMAL_ARRAY);

        gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
    }

    /**
     * Restore all openGL state.
     *
      * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
    }

    /**
     * Check for all intersections against this geometry using a line segment and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param start The start point of the segment
     * @param end The end point of the segment
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public boolean pickLineSegment(float[] start,
                                   float[] end,
                                   boolean findAny,
                                   float[] dataOut,
                                   int dataOutFlags)
        throws NotPickableException
    {
        // Call the super version to do our basic checking for us. Ignore the
        // return value as we're about to go calculate that ourselves.
        super.pickLineSegment(start, end, findAny, dataOut, dataOutFlags);

        float shortest_length = Float.POSITIVE_INFINITY;
        float this_length;
        boolean found = false;
        float out_x = 0;
        float out_y = 0;
        float out_z = 0;

        // copy the end var to a local for the time being as we're going to
        // reuse end as a direction vector.
        float end_x = end[0];
        float end_y = end[1];
        float end_z = end[2];

        float x = end[0] - start[0];
        float y = end[1] - start[1];
        float z = end[2] - start[2];

        float vec_len = (float)Math.sqrt(x * x + y * y + z * z);
        end[0] = x;
        end[1] = y;
        end[2] = z;

        int coord_offset = 0;

        for(int i = 0; i < numFans; i++)
        {
            int num_tris = fanCounts[i] - 2;
            wkPolygon[0] = coordinates[coord_offset++];
            wkPolygon[1] = coordinates[coord_offset++];
            wkPolygon[2] = coordinates[coord_offset++];

            for(int j = 0; j < num_tris; j++)
            {
                wkPolygon[3] = coordinates[coord_offset++];
                wkPolygon[4] = coordinates[coord_offset++];
                wkPolygon[5] = coordinates[coord_offset++];

                wkPolygon[6] = coordinates[coord_offset];
                wkPolygon[7] = coordinates[coord_offset + 1];
                wkPolygon[8] = coordinates[coord_offset + 2];

                if(ray3DTriangleChecked(start, end, vec_len, dataOut))
                {
                    found = true;

                    if(findAny)
                        break;

                    float l_x = start[0] - dataOut[0];
                    float l_y = start[1] - dataOut[1];
                    float l_z = start[2] - dataOut[2];

                    this_length = l_x * l_x + l_y * l_y + l_z * l_z;

                    if(this_length < shortest_length)
                    {
                        shortest_length = this_length;
                        out_x = dataOut[0];
                        out_y = dataOut[1];
                        out_z = dataOut[2];
                    }
                }
            }

            // Shift to the start of the next fan
            coord_offset += 3;
        }

        dataOut[0] = out_x;
        dataOut[1] = out_y;
        dataOut[2] = out_z;

        // Copy it back again.
        end[0] = end_x;
        end[1] = end_y;
        end[2] = end_z;

        return found;
    }

    /**
     * Check for all intersections against this geometry using a line ray and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param origin The start point of the ray
     * @param direction The direction vector of the ray
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public boolean pickLineRay(float[] origin,
                               float[] direction,
                               boolean findAny,
                               float[] dataOut,
                               int dataOutFlags)
        throws NotPickableException
    {
        // Call the super version to do our basic checking for us. Ignore the
        // return value as we're about to go calculate that ourselves.
        super.pickLineRay(origin, direction, findAny, dataOut, dataOutFlags);

        float shortest_length = Float.POSITIVE_INFINITY;
        float this_length;
        boolean found = false;
        float out_x = 0;
        float out_y = 0;
        float out_z = 0;

        int coord_offset = 0;

        for(int i = 0; i < numFans; i++)
        {
            int num_tris = fanCounts[i] - 2;
            wkPolygon[0] = coordinates[coord_offset++];
            wkPolygon[1] = coordinates[coord_offset++];
            wkPolygon[2] = coordinates[coord_offset++];

            for(int j = 0; j < num_tris; j++)
            {
                wkPolygon[3] = coordinates[coord_offset++];
                wkPolygon[4] = coordinates[coord_offset++];
                wkPolygon[5] = coordinates[coord_offset++];

                wkPolygon[6] = coordinates[coord_offset];
                wkPolygon[7] = coordinates[coord_offset + 1];
                wkPolygon[8] = coordinates[coord_offset + 2];

                if(ray3DTriangleChecked(origin, direction, 0, dataOut))
                {
                    found = true;

                    if(findAny)
                        break;

                    float l_x = origin[0] - dataOut[0];
                    float l_y = origin[1] - dataOut[1];
                    float l_z = origin[2] - dataOut[2];

                    this_length = l_x * l_x + l_y * l_y + l_z * l_z;

                    if(this_length < shortest_length)
                    {
                        shortest_length = this_length;
                        out_x = dataOut[0];
                        out_y = dataOut[1];
                        out_z = dataOut[2];
                    }
                }
            }

            // Shift to the start of the next fan
            coord_offset += 3;
        }

        dataOut[0] = out_x;
        dataOut[1] = out_y;
        dataOut[2] = out_z;

        return found;
    }
}