/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
// None

// Local imports
import org.j3d.aviatrix3d.DrawableSurface;
import org.j3d.aviatrix3d.RenderPipeline;
import org.j3d.aviatrix3d.DeletableSceneGraphObject;

/**
 * Wrapper around a single pipeline that controls its rendering in a separate
 * thread.
 *
 * @author Justin Couch
 * @version $Revision: 1.3 $
 */
class PipelineThread extends Thread
{
    /** flag to tell the thread to exit now */
    private boolean terminate;

    /** The pipeline we're managing */
    private RenderPipeline pipeline;

    /** Callback to tell the system we've finished this render */
    private PipelineStateObserver observer;

    /** Lock object used to control when to run the display */
    private Object displayLock;

    /** Indicate that this is waiting for rendering to complete */
    private boolean waiting;

    /** Flag to decide whether we should only display the last frame */
    private boolean noProcessing;

    /**
     * Construct an instance that manages the given pipeline.
     */
    PipelineThread(RenderPipeline rp)
    {
        pipeline = rp;
        terminate = false;
        waiting = false;
        noProcessing = false;
        displayLock = new Object();
    }

    //---------------------------------------------------------------
    // Methods defined by Runnable
    //---------------------------------------------------------------

    /**
     * Endless loop that does the display calls asynchronously.
     */
    public void run()
    {
        DrawableSurface drawable = pipeline.getDrawableSurface();
        drawable.setRenderControlThread(this);

        while(!terminate)
        {
            try
            {
                if(!terminate)
                {
                    if(noProcessing)
                        pipeline.displayOnly();
                    else
                        pipeline.render();
                }

                synchronized(displayLock)
                {
                    if(observer != null)
                        observer.frameFinished();

                    displayLock.wait();
                }
            }
            catch(InterruptedException ie)
            {
            }
        }

        drawable.setRenderControlThread(null);
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Instruct the pipeline to only display and not do any processing.
     */
    void displayOnly()
    {
        noProcessing = true;
    }

    /**
     * Request the contained pipeline renders now. If the pipeline is currently
     * processing a previous display request, this is ignored.
     */
    void render()
    {
        synchronized(displayLock)
        {
            displayLock.notify();
        }
    }

    /**
     * Force a shutdown of the pipeline now. If the display is currently
     * processing, the current processing finishes before the thread
     * terminates.
     */
    void shutdown()
    {
        terminate = true;
        synchronized(displayLock)
        {
            displayLock.notify();
        }
    }

    /**
     * Set the observer instance used to watch for when this pipeline has
     * finished rendering.
     *
     * @param pso
     */
    void setStateObserver(PipelineStateObserver pso)
    {
        observer = pso;
    }

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param queue The instance of the queue to use or null
     */
    public void queueDeletedObject(DeletableSceneGraphObject obj)
    {
        DrawableSurface surf = pipeline.getDrawableSurface();
        surf.queueDeletedObject(obj);
    }
}