/*****************************************************************************
 *                        Web3d.org Copyright (c) 2004
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Point4f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Handles the scenegraph maintenance and culling operations.
 * <p>
 *
 * The culling phase generates a list of nodes to render.
 * A future optimization will sort the render list by OGL state.
 * <p>
 *
 * By default, the implementation will walk into shape3Ds looking for any
 * offscreen textures to be rendered. If you know that you do not have any
 * in the scene, then you can set an internal flag to not look for them, thus
 * achieving a simple performance boost.
 *
 * @author Alan Hudson
 * @version $Revision: 1.21 $
 */
public class SimpleFrustumCullStage implements CullStage
{
    /** Message when we discover the VP is under a shared group/node */
    private static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup instance. This is not permitted.";

    /** Message when we discover the VP is under a shared group/node */
    private static final String SHARED_BG_MSG =
        "The path from the root of the scene graph to the background contains " +
        "a SharedGroup instance. This is not permitted.";

    /** Initial depth for the number of transforms above a view,fog or bg */
    private static final int INITIAL_PATH_DEPTH = 64;

    /** Initial scene depth used for the transform stack */
    private static final int TRANSFORM_DEPTH_SIZE = 64;

    /** The increment size of the stack if it gets overflowed */
    private static final int STACK_INCREMENT = 32;

    /** The initial size of the children list */
    private static final int LIST_START_DEPTH = 200;

    /** Initial size of the cull list for the number of offscreen surfaces */
    private static final int LIST_START_LENGTH = 1;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 20;

    /** The initial size of the light list */
    private static final int LIGHT_START_SIZE = 32;

    /** The increment size of the list if it gets overflowed */
    private static final int LIGHT_INCREMENT = 8;

    /** If set, recurse into Shape3Ds looking for offscreen textures */
    private boolean checkOffscreens;

    /** Buffer list of sub scenes that still need processing */
    private OffscreenTextureSource[] pendingSubscenes;

    /** Buffer list of sub scenes that still need processing */
    private OffscreenTextureSource[] pendingParents;

    /** Index of the last subscene in the list */
    private int lastSubscene;

    /**
     * List of final subscenes to send to the final stage. First index is
     * the direct owner of the scene contents. The second index is the
     * scene parent of the scene included (needed for pBuffer GL context
     * handling at render time). If this second one is null, then the parent
     * is the main canvas that is being rendered to.
     */
    private SceneGraphObject[][] outputSceneParents;

    /** List of valid subscene parents as we process them */
    private SceneGraphObject[][] validSceneParents;

    /** The list of culled nodes to send to the next stage */
    private CullOutputDetails[][] outputCullList;

    /** List of valid culled nodes as we process them */
    private CullOutputDetails[][] validCullList;

    /** List that is being used to fill values into */
    private CullOutputDetails[] workCullList;

    /** list of culled nodes list sizes to send to the next stage */
    private int[] outputCullSize;

    /** List of environment data as we process them */
    private RenderEnvironmentData[] validEnvData;

    /** list of environment data to send to the next stage */
    private RenderEnvironmentData[] outputEnvData;

    /** List of culled nodes list sizes as we process them */
    private int[] validCullSize;

    /** Index to the next output list in the validCullList */
    private int lastOutputList;

    /** A stack used to control the depth of the transform tree */
    private Matrix4f[] transformStack;

    /** Index to the next place to add items in the transformStack */
    private int lastStack;

    /** The list of lights nodes currently valid while traversing. */
    private Light[] lightList;

    /** The list of light transforms currently valid while traversing */
    private float[][] transformList;

    /** Index to the next place to add items in the nodeList */
    private int lastLight;

    /** Temp used to get texture units when looking for offscreen textures. */
    private TextureUnit[] textureUnits;

    /** Collection of offscreen textures we've found this frame */
    private HashSet processedOffscreenTextures;

    /** Collection of textures we've found for the current parent texture */
    private HashSet visitedNodes;

    /** Handler for the output */
    private CulledGeometryReceiver receiver;

    /** The active parent scene of the scene we are processing. */
    private OffscreenTextureSource activeParent;

    /** Working var for the view frustum */
    private double[] viewFrustum;

    /** Working var to calculate the view matrix */
    private Matrix4f viewMatrix;

    /** The 8 bounding points of the frustum volume */
    private Point4f[] frustumPoints;

    /** The planes describing this frustum */
    private Vector4f[] frustumPlanes;

    /** Projection matrix used to generate the frustum planes with. */
    private Matrix4f prjMatrix;

    /** Path to the current viewpoint */
    private TransformGroup[] currentViewpointPath;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix1;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix2;

    /** Rotation matrix for the orientation provided by user */
    private Matrix4f orientationMatrix;

    /** Storage variable for the screen orientation values */
    private float[] screenOrientation;

    /** Storage variable for the eye offset values */
    private float[] eyePoint;

    /** Flag to say explicit screen orientation values have been provided */
    private boolean useOrientation;

    /** Flag to say explicit eyepoint values have been provided */
    private boolean useEyePoint;

    int cullWait = 10;
    int reject;
    int keep;
    int level;

    long startTime;

    /**
     * Create a basic instance of this class with the list assuming there are
     * no off-screen buffers in use for the initial internal setup.
     */
    public SimpleFrustumCullStage()
    {
        this(LIST_START_LENGTH);
    }

    /**
     * Create a basic instance of this class with the list initial internal
     * setup for the given number of renderable surfaces. The size is just an
     * initial esstimate, and is used for optimisation purposes to prevent
     * frequent array reallocations internally. As such, the number does not
     * have to be perfect, just good enough.
     *
     * @param numSurfaces Total number of surfaces to prepare rendering for
     */
    public SimpleFrustumCullStage(int numSurfaces)
    {
        outputCullList = new CullOutputDetails[numSurfaces][];
        validCullList = new CullOutputDetails[numSurfaces][LIST_START_DEPTH];
        outputSceneParents = new SceneGraphObject[numSurfaces][2];
        validSceneParents = new SceneGraphObject[numSurfaces][2];
        pendingSubscenes = new OffscreenTextureSource[numSurfaces];
        pendingParents = new OffscreenTextureSource[numSurfaces];
        outputEnvData = new RenderEnvironmentData[numSurfaces];
        validEnvData = new RenderEnvironmentData[numSurfaces];
        outputCullSize = new int[numSurfaces];
        validCullSize = new int[numSurfaces];

        textureUnits = new TextureUnit[32];
        processedOffscreenTextures = new HashSet();
        visitedNodes = new HashSet();

        orientationMatrix = new Matrix4f();
        orientationMatrix.setIdentity();

        screenOrientation = new float[4];
        eyePoint = new float[3];

        useEyePoint = false;
        useOrientation = false;

        // populate the list with valid instances...
        for(int i = 0; i < numSurfaces; i++)
            validEnvData[i] = new RenderEnvironmentData();

        // populate the list with valid instances...
        for(int i = 0; i < numSurfaces; i++)
            for(int j = 0; j < LIST_START_DEPTH; j++)
                validCullList[i][j] = new CullOutputDetails();

        transformStack = new Matrix4f[TRANSFORM_DEPTH_SIZE];

        // populate the list with valid instances...
        for(int i = 0; i < TRANSFORM_DEPTH_SIZE; i++)
            transformStack[i] = new Matrix4f();

        lightList = new Light[LIGHT_START_SIZE];
        transformList = new float[LIGHT_START_SIZE][16];

        checkOffscreens = true;
        prjMatrix = new Matrix4f();
        frustumPoints = new Point4f[8];
        for(int i=0; i < 8; i++)
            frustumPoints[i] = new Point4f();

        frustumPlanes = new Vector4f[6];
        for(int i=0; i < 6; i++)
            frustumPlanes[i] = new Vector4f();

        currentViewpointPath = new TransformGroup[INITIAL_PATH_DEPTH];
        viewMatrix = new Matrix4f();
        viewMatrix1 = new Matrix4f();
        viewMatrix2 = new Matrix4f();
    }

    //---------------------------------------------------------------
    // Methods defined by CullStage
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param scene The scene instance to cull
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Scene scene, int buffer)
    {
        // Buffer ID is not used currently.
        startTime = System.currentTimeMillis();

        reject = 0;
        keep = 0;

        if(cullWait > 0)
            cullWait--;

        lastOutputList = 0;
        lastLight = 0;
        lastStack = 0;
        lastSubscene = 0;
        activeParent = null;

        workCullList = validCullList[0];
        validSceneParents[0][0] = null;
        validSceneParents[0][1] = null;

        Node node = scene.getRenderedGeometry();
        fillEnvData(scene, validEnvData[0]);


        if(node instanceof Group)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();

            if (node instanceof TransformGroup) {
                ((TransformGroup)node).getTransform(transformStack[0]);
                // TODO: shouldn't this increment lastStack?
            }

            viewFrustum = validEnvData[0].viewFrustum;
            viewMatrix.set(validEnvData[0].viewTransform);

            updateFrustum(validEnvData[0]);

//System.out.println("Culling:");
            validCullSize[0] = findAllNodes((Group)node, false, false, 0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if(node instanceof SharedNode)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();

            viewFrustum = validEnvData[0].viewFrustum;
            viewMatrix.set(validEnvData[0].viewTransform);

            updateFrustum(validEnvData[0]);

            validCullSize[0] = findNextNode((SharedNode)node, false, false, 0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if((node instanceof Shape3D) && ((Shape3D)node).isVisible())
        {
            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)node);

            workCullList[0].renderable = (Leaf)node;

            // Manual set of the stack back to an identity matrix
            workCullList[0].transform.setIdentity();
            workCullList[0].numLights = 0;

            validCullSize[0] = 1;
            validCullList[0] = workCullList;
            lastOutputList++;
        }


        if(checkOffscreens)
        {
            // May need to shift inside the loop for cases where textures
            // contain scenes which contain further textures. Play a wait and
            // see game for now.
            resizeOutputLists();

            for(int i = 0; i < lastSubscene; i++)
            {
                activeParent = pendingSubscenes[i];
                visitedNodes.clear();

                if(processedOffscreenTextures.contains(pendingSubscenes[i]))
                {
                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];
                }
                else
                {
                    Scene sub_scene = pendingSubscenes[i].getRenderedScene();

                    Group grp = sub_scene.getRenderedGeometry();

                    if(grp == null)
                        continue;

                    processedOffscreenTextures.add(pendingSubscenes[i]);

                    // Reset these for each separate scene
                    lastLight = 0;
                    lastStack = 0;

                    // Since the first output cull list index is used by the
                    workCullList = validCullList[lastOutputList];

                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];

                    fillEnvData(sub_scene, validEnvData[lastOutputList]);

                    validCullSize[lastOutputList] = cullSingle(grp);

                    // Assign it back, just in case the list got resized.
                    validCullList[lastOutputList] = workCullList;
                }

                lastOutputList++;
            }

            // Now reverse the list since we're doing a breadth-first search
            // for the subscenes, we end up with the most nested subscene last.
            // This is bad. So now reverse the order of the list so that the
            // most nested collection is first to be rendered and the main
            // screen is the last item.
            if(outputCullList.length < lastOutputList)
            {
                outputCullList = new CullOutputDetails[lastOutputList][];
                outputCullSize = new int[lastOutputList];
                outputEnvData = new RenderEnvironmentData[lastOutputList];
                outputSceneParents = new SceneGraphObject[lastOutputList][2];
            }

            for(int i = 0; i < lastOutputList; i++)
            {
                outputCullList[i] = validCullList[lastOutputList - i - 1];
                outputCullSize[i] = validCullSize[lastOutputList - i - 1];
                outputEnvData[i]  = validEnvData[lastOutputList - i - 1];
                outputSceneParents[i][0] =
                    validSceneParents[lastOutputList - i - 1][0];
                outputSceneParents[i][1] =
                    validSceneParents[lastOutputList - i - 1][1];
            }
        }
        else
        {
            // It is likely in this case that we're doing single-scene
            // rendering and will be doing it all the time. Possibly optimise
            // this by avoiding the copy altogether, but not 100% this won't
            // break in some corner case right now.

            outputCullList[0] = validCullList[0];
            outputCullSize[0] = validCullSize[0];
            outputEnvData[0]  = validEnvData[0];
            outputSceneParents[0][0] = validSceneParents[0][0];
            outputSceneParents[0][1] = validSceneParents[0][1];
        }

        if((lastOutputList != 0) && (receiver != null))
            receiver.culledOutput(outputEnvData,
                                  outputCullList,
                                  outputCullSize,
                                  outputSceneParents,
                                  lastOutputList);

        processedOffscreenTextures.clear();

//System.out.println("size: " + outputCullSize[0] + " Keep: " + keep + " reject: " + reject + " error: " + errorBB + " time: " + (System.currentTimeMillis() - startTime));
    }

    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setCulledGeometryReceiver(CulledGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Set the flag for whether to check for offscreen textures or not. By
     * default, this flag is set to true.
     *
     * @param state true if offscreen textures should be looked for
     */
    public void setOffscreenCheckEnabled(boolean state)
    {
        checkOffscreens = state;
    }

    /**
     * Find out what the current offscreen check state is.
     *
     * @return true if the checking is being performed
     */
    public boolean isOffscreenCheckEnabled()
    {
        return checkOffscreens;
    }

    /**
     * Set the eyepoint offset from the centre position. This is used to model
     * offset view frustums, such as multiple displays or a powerwall. This
     * method will be called with the appropriate values from the
     * RenderPipeline that this culler is inserted into.
     *
     * @param x The x axis offset
     * @param y The y axis offset
     * @param z The z axis offset
     */
    public void setEyePointOffset(float x, float y, float z)
    {
        useEyePoint = true;
        eyePoint[0] = x;
        eyePoint[1] = y;
        eyePoint[2] = z;
    }

    /**
     * Set the orientation of this screen relative to the user's normal view
     * direction. The normal orientation of the screen is along the negative
     * Z axis. This method provides and axis-angle reorientation of that
     * direction to one that is facing the screen. Typically this will just
     * involve a rotation around the Y axis of some amount (45 and 90 deg
     * being the most common used in walls and caves). This method will be
     * called with the appropriate values from the RenderPipeline that this
     * culler is inserted into.
     *
     * @param x The x axis component
     * @param y The y axis component
     * @param z The z axis component
     * @param a The angle to rotate around the axis in radians
     * @throws IllegalArgumentException The length of the axis is zero
     */
    public void setScreenOrientation(float x, float y, float z, float a)
      throws IllegalArgumentException
    {
        float n = x * x + y * y + z * z;
        if(n == 0)
            throw new IllegalArgumentException("Zero length axis vector");

        useOrientation = true;
        screenOrientation[0] = x;
        screenOrientation[1] = y;
        screenOrientation[2] = z;
        screenOrientation[3] = a;

        double rcos = Math.cos(a);
        double rsin = Math.sin(a);

        orientationMatrix.m00 = (float)(rcos + x * x * (1 - rcos));
        orientationMatrix.m01 = (float)(z * rsin + y * x * (1 - rcos));
        orientationMatrix.m02 = (float)(-y * rsin + z * x * (1 - rcos));
        orientationMatrix.m10 = (float)(-z * rsin + x * y * (1 - rcos));
        orientationMatrix.m11 = (float)(rcos + y * y * (1 - rcos));
        orientationMatrix.m12 = (float)(x * rsin + z * y * (1 - rcos));
        orientationMatrix.m20 = (float)(y * rsin + x * z * (1 - rcos));
        orientationMatrix.m21 = (float)(-x * rsin + y * z * (1 - rcos));
        orientationMatrix.m22 = (float)(rcos + z * z * (1 - rcos));
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph of a single node without doing any other
     * processing. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param node The root node to start the cull from
     */
    private int cullSingle(Group node)
    {
        // set the top matrix stack back to identity
        transformStack[0].setIdentity();

        if (node instanceof TransformGroup) {
            ((TransformGroup)node).getTransform(transformStack[0]);
        }

        return findAllNodes((Group)node, false, false, 0);
    }

    /**
     * Recursive walk of the tree to find all the renderable nodes.
     *
     * @param group The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @param allInBounds All of this group is inside the view frustum
     * @param ignoreTrans Ignore the last transformation
     * @return The index of the last item on the cull list
     */
    private int findAllNodes(Group group,
                             boolean allInBounds,
                             boolean ignoreTrans,
                             int cullEndIndex)
    {

        BoundingVolume bounds = group.getBounds();
        boolean childAllInBounds = false;

//System.out.println(levelTab() + group + " allIn: " + allInBounds + " bounds: " + bounds);
//level++;


        if(!allInBounds)
        {
            Matrix4f mat;

            int matIdx;

            if(ignoreTrans)
                matIdx = lastStack-1;
            else
                matIdx = lastStack;

            mat = transformStack[matIdx];

            int result = bounds.checkIntersectionFrustum(frustumPlanes, mat);
/*
if(bounds instanceof BoundingVoid) {
    result = BoundingVolume.FRUSTUM_PARTIAL;
}
*/
            switch(result)
            {
                case BoundingVolume.FRUSTUM_ALLOUT:
                    reject++;
                    level--;
                    return cullEndIndex;
                case BoundingVolume.FRUSTUM_ALLIN:
                    keep++;
                    childAllInBounds = true;
                    break;
                case BoundingVolume.FRUSTUM_PARTIAL:
                    keep++;
                    break;
            }
        }

        Node[] kids  = group.getRenderableChild();
        int size = group.numRenderableChildren();

        int num_lights = 0;
        int ret_val = cullEndIndex;

        // Find all the lights at this level first. Place them first into
        // the queue
        for(int i = 0; i < size; i++)
        {
            if(!(kids[i] instanceof Light))
                continue;

            Light l = (Light)kids[i];
            if(!l.isEnabled())
                continue;

            resizeLightList();
            lightList[lastLight] = l;

            Matrix4f mat = transformStack[lastStack];

            // Transpose the matrix in place as it is being copied
            transformList[lastLight][0] = (float)mat.m00;
            transformList[lastLight][1] = (float)mat.m10;
            transformList[lastLight][2] = (float)mat.m20;
            transformList[lastLight][3] = (float)mat.m30;

            transformList[lastLight][4] = (float)mat.m01;
            transformList[lastLight][5] = (float)mat.m11;
            transformList[lastLight][6] = (float)mat.m21;
            transformList[lastLight][7] = (float)mat.m31;

            transformList[lastLight][8] = (float)mat.m02;
            transformList[lastLight][9] = (float)mat.m12;
            transformList[lastLight][10] = (float)mat.m22;
            transformList[lastLight][11] = (float)mat.m32;

            transformList[lastLight][12] = (float)mat.m03;
            transformList[lastLight][13] = (float)mat.m13;
            transformList[lastLight][14] = (float)mat.m23;
            transformList[lastLight][15] = (float)mat.m33;

            lastLight++;
            num_lights++;
        }



        for(int i = 0; i < size; i++)
        {
            if(kids[i] instanceof Group)
            {
                // If a tg, push the new TX onto the stack
                boolean is_tx = (kids[i] instanceof TransformGroup);

                if(is_tx)
                {
                    resizeStack();
                    TransformGroup tg = (TransformGroup)kids[i];
                    tg.getTransform(transformStack[lastStack + 1]);

                    transformStack[lastStack + 1].mul(transformStack[lastStack],
                                                      transformStack[lastStack + 1]);
                    lastStack++;
                    ret_val = findAllNodes((Group)kids[i],
                                           childAllInBounds,
                                           true,
                                           ret_val);
                }
                else
                {
                    ret_val = findAllNodes((Group)kids[i],
                                           childAllInBounds,
                                           false,
                                           ret_val);
                }

                // Now pop the stack.
                if(is_tx)
                    lastStack--;
            }
            else if(kids[i] instanceof Shape3D)
            {
                // Check the visibility state and ignore if not visible.
                if(!((Shape3D)kids[i]).isVisible())
                    continue;

                // Walk into the shape and check that we don't have any
                // offscreen textures to render.
                if(checkOffscreens)
                    checkForOffscreens((Shape3D)kids[i]);

                resizeCullList(ret_val);
                workCullList[ret_val].renderable = (Leaf)kids[i];

                Matrix4f mat = transformStack[lastStack];

                workCullList[ret_val].transform.set(mat);

                int src_size = (workCullList[ret_val].lights == null) ?
                                0 :
                                workCullList[ret_val].lights.length;

                if(src_size < lastLight)
                {
                    // up the size of the array

                    LightDetails[] tmp = new LightDetails[lastLight];
                    if(src_size != 0)
                    {
                        System.arraycopy(workCullList[ret_val].lights,
                                         0,
                                         tmp,
                                         0,
                                         src_size);
                    }

                    for(int j = src_size; j < lastLight; j++)
                        tmp[j] = new LightDetails();

                    workCullList[ret_val].lights = tmp;
                }

                // copy in the light information.
                LightDetails[] tmp = workCullList[ret_val].lights;
                workCullList[ret_val].numLights = lastLight;

                for(int j = 0; j < lastLight; j++)
                    tmp[j].update(lightList[j], transformList[j]);

                ret_val++;
            }
            else if(kids[i] instanceof SharedNode)
            {
                ret_val = findNextNode((SharedNode)kids[i],
                                       true,
                                       false,
                                       ret_val);
            }
        }

        // Pop the valid lights from the stack
        lastLight -= num_lights;

//level--;
        return ret_val;
    }

    /**
     * From the given SharedNode instance keep finding a non-sharedNode
     * instance, before continuing on the recursion
     *
     * @param root The parent node to walk into
     * @param allInBounds All of this group is inside the view frustum
     * @param ignoreTrans Ignore the last transformation
     * @param cullEndIndex The current last item on the cull list
     * @return The index of the last item on the cull list
     */
    private int findNextNode(SharedNode root,
                             boolean allInBounds,
                             boolean ignoreTrans,
                             int cullEndIndex)
    {
        int ret_val = cullEndIndex;

        Node current = root.getChild();
//System.out.println(levelTab() + root + " allIn: " + allInBounds);

        while(current != null && current instanceof SharedNode)
            current = ((SharedNode)current).getChild();

        // Now just continue on normally. If the result of the above was a
        // null then it just falls off the end of this if/else naturally.
        if(current instanceof Group)
        {
            // If a tg, push the new TX onto the stack
            boolean is_tx = (current instanceof TransformGroup);

            if(is_tx)
            {
                resizeStack();
                TransformGroup tg = (TransformGroup)current;
                tg.getTransform(transformStack[lastStack + 1]);

                transformStack[lastStack + 1].mul(transformStack[lastStack],
                                                  transformStack[lastStack + 1]);
                lastStack++;

                ret_val = findAllNodes((Group)current,
                                       allInBounds,
                                       true,
                                       ret_val);
            }
            else
            {
                ret_val = findAllNodes((Group)current,
                                       allInBounds,
                                       false,
                                       ret_val);
            }

            // Now pop the stack.
            if(is_tx)
                lastStack--;
        }
        else if(current instanceof Shape3D)
        {
            // Check the visibility state and ignore if not visible.
            if(!((Shape3D)current).isVisible())
                return ret_val;

            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)current);

            resizeCullList(cullEndIndex);
            workCullList[cullEndIndex].renderable = (Leaf)current;

            Matrix4f mat = transformStack[lastStack];

            // Transpose the matrix in place as it is being copied
            workCullList[cullEndIndex].transform.set(mat);

            int src_size = (workCullList[cullEndIndex].lights == null) ?
                            0 :
                            workCullList[cullEndIndex].lights.length;

            if(src_size < lastLight)
            {
                // up the size of the array

                LightDetails[] tmp = new LightDetails[lastLight];
                if(src_size != 0)
                {
                    System.arraycopy(workCullList[cullEndIndex].lights,
                                     0,
                                     tmp,
                                     0,
                                     src_size);
                }

                for(int j = src_size; j < lastLight; j++)
                    tmp[j] = new LightDetails();

                workCullList[cullEndIndex].lights = tmp;
            }

            // copy in the light information.
            LightDetails[] tmp = workCullList[cullEndIndex].lights;
            workCullList[cullEndIndex].numLights = lastLight;

            for(int j = 0; j < lastLight; j++)
                tmp[j].update(lightList[j], transformList[j]);

            cullEndIndex++;

            ret_val++;
        }

        return ret_val;
    }

    /**
     * Take the scene and fill in a RenderEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     */
    private void fillEnvData(Scene scene, RenderEnvironmentData envData)
    {
        ViewEnvironment view = scene.getViewEnvironment();

        envData.userData = scene.getUserData();
        envData.effectsProcessor = scene.getRenderEffectsProcessor();
        envData.viewpoint = scene.getActiveView();
        envData.fog = scene.getActiveFog();
        envData.background = scene.getActiveBackground();
        envData.viewEnvironment = view;

        envData.eyeOffset[0] = eyePoint[0];
        envData.eyeOffset[1] = eyePoint[1];
        envData.eyeOffset[2] = eyePoint[2];

        // Calculate frustum parameters based on view environment
        double xmin, xmax, ymin, ymax;
        double aspect, fov;
        double zNear, zFar;

        if(view == null)
        {
            aspect = 1;
            zNear = 0.01;
            zFar = 1000;
            fov = 45;
        }
        else
        {
            aspect = view.getAspectRatio();
            zNear = view.getNearClipDistance();
            zFar = view.getFarClipDistance();
            fov = view.getFieldOfView();
        }

        ymax = zNear * Math.tan(fov * Math.PI / 360.0);
        ymin = -ymax;
        xmin = ymin * aspect;
        xmax = ymax * aspect;

        if(useEyePoint)
        {
            envData.viewFrustum[0] = xmin - eyePoint[0];
            envData.viewFrustum[1] = xmax - eyePoint[0];
            envData.viewFrustum[2] = ymin - eyePoint[1];
            envData.viewFrustum[3] = ymax - eyePoint[1];
            envData.viewFrustum[4] = zNear - eyePoint[2];
            envData.viewFrustum[5] = zFar - eyePoint[2];
        }
        else
        {
            envData.viewFrustum[0] = xmin;
            envData.viewFrustum[1] = xmax;
            envData.viewFrustum[2] = ymin;
            envData.viewFrustum[3] = ymax;
            envData.viewFrustum[4] = zNear;
            envData.viewFrustum[5] = zFar;
        }


        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = envData.viewpoint.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];

        // start at the VP and work our way upwards.
        parent = envData.viewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }

        if(useOrientation)
            viewMatrix2.mul(orientationMatrix);

        envData.viewTransform.set(viewMatrix2);

        // Now do almost the same thing with the background node. Only
        // difference is that this time we deliberately set the translation
        // components to zero.
        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        if(envData.background != null)
        {
            parent = envData.background.getParent();
            path_depth = 0;

            while(parent != null)
            {
                if((parent instanceof SharedGroup) ||
                   (parent instanceof SharedNode))
                    throw new IllegalArgumentException(SHARED_BG_MSG);

                parent = parent.getParent();
                path_depth++;
            }

            // reuse the VP path for this one
            if(currentViewpointPath.length < path_depth)
                currentViewpointPath = new TransformGroup[path_depth];

            // start at the background and work our way upwards.
            parent = envData.background.getParent();
            last_tx = 0;
            while(parent != null)
            {
                if(parent instanceof TransformGroup)
                    currentViewpointPath[last_tx++] = (TransformGroup)parent;

                parent = parent.getParent();
            }

            viewMatrix2.setIdentity();

            if(last_tx != 0)
            {
                for(int i = last_tx - 1; i >= 0; i--)
                {
                    currentViewpointPath[i].getTransform(viewMatrix1);
                    viewMatrix2.mul(viewMatrix1);
                }
            }

            Matrix4f bg_mat = envData.backgroundTransform;

            bg_mat.m00 = viewMatrix2.m00;
            bg_mat.m01 = viewMatrix2.m01;
            bg_mat.m02 = viewMatrix2.m02;
            bg_mat.m03 = viewMatrix2.m03;
            bg_mat.m10 = viewMatrix2.m10;
            bg_mat.m11 = viewMatrix2.m11;
            bg_mat.m12 = viewMatrix2.m12;
            bg_mat.m13 = viewMatrix2.m13;
            bg_mat.m20 = viewMatrix2.m20;
            bg_mat.m21 = viewMatrix2.m21;
            bg_mat.m22 = viewMatrix2.m22;
            bg_mat.m23 = viewMatrix2.m23;
            bg_mat.m30 = 0;
            bg_mat.m31 = 0;
            bg_mat.m32 = 0;
            bg_mat.m33 = viewMatrix2.m33;
        }
    }

    /**
     * Check a shape node for the offscreen textures that may be present. If
     * some are found, queue them up for processing.
     *
     * @param shape The object instance to process
     */
    private void checkForOffscreens(Shape3D shape)
    {
        Appearance app = shape.getAppearance();

        if(app == null)
            return;

        int num_units = app.numTextureUnits();
        if(num_units == 0)
            return;

        app.getTextureUnits(textureUnits);

        for(int i =0; i < num_units; i++)
        {
            Texture tex = textureUnits[i].getTexture();

            // Only queue up the texture for repainting if it says it needs a
            // repaint and that we haven't already found this instance in this
            // frame.
            if((tex instanceof OffscreenTextureSource) &&
               !visitedNodes.contains(tex) &&
               ((OffscreenTextureSource)tex).isRepaintRequired())
            {
                resizeOffscreenList();
                pendingSubscenes[lastSubscene] = (OffscreenTextureSource)tex;
                pendingParents[lastSubscene] = activeParent;
                visitedNodes.add(tex);
                lastSubscene++;
            }
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeCullList(int cur_size)
    {
        if((cur_size + 1) == workCullList.length)
        {
            int old_size = workCullList.length;
            int new_size = old_size + LIST_INCREMENT;

            CullOutputDetails[] tmp_nodes = new CullOutputDetails[new_size];

            System.arraycopy(workCullList, 0, tmp_nodes, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_nodes[i] = new CullOutputDetails();

            workCullList = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeOutputLists()
    {
        if((lastSubscene + 1) < validCullList.length)
            return;

        int old_size = validCullList.length;
        int new_size = old_size + LIST_INCREMENT;

        CullOutputDetails[][] tmp_nodes =
            new CullOutputDetails[new_size][];

        System.arraycopy(validCullList, 0, tmp_nodes, 0, old_size);

        for(int i = old_size; i < new_size; i++)
        {
            tmp_nodes[i] = new CullOutputDetails[LIST_START_DEPTH];

            for(int j = 0; j < LIST_START_DEPTH; j++)
                tmp_nodes[i][j] = new CullOutputDetails();
        }

        validCullList = tmp_nodes;

        int[] tmp_ints = new int[new_size];
        System.arraycopy(validCullSize, 0, tmp_ints, 0, old_size);

        validCullSize = tmp_ints;

        SceneGraphObject[][] tmp_obj = new SceneGraphObject[new_size][];

        System.arraycopy(validSceneParents, 0, tmp_obj, 0, old_size);

        for(int i = old_size; i < new_size; i++)
            tmp_obj[i] = new SceneGraphObject[2];

        validSceneParents = tmp_obj;

        RenderEnvironmentData[] tmp_env = new RenderEnvironmentData[new_size];
        System.arraycopy(validEnvData, 0, tmp_env, 0, old_size);

         for(int i = old_size; i < new_size; i++)
             tmp_env[i] = new RenderEnvironmentData();

        validEnvData = tmp_env;
    }


    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeOffscreenList()
    {
        if((lastSubscene + 1) == pendingSubscenes.length)
        {
            int old_size = pendingSubscenes.length;
            int new_size = old_size + LIST_INCREMENT;

            OffscreenTextureSource[] tmp_nodes =
                new OffscreenTextureSource[new_size];

            System.arraycopy(pendingSubscenes, 0, tmp_nodes, 0, old_size);

            pendingSubscenes = tmp_nodes;

            tmp_nodes = new OffscreenTextureSource[new_size];

            System.arraycopy(pendingParents, 0, tmp_nodes, 0, old_size);

            pendingParents = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeLightList()
    {
        if((lastLight + 1) == lightList.length)
        {
            int old_size = lightList.length;
            int new_size = old_size + LIST_INCREMENT;

            Light[] tmp_nodes = new Light[new_size];

            System.arraycopy(lightList, 0, tmp_nodes, 0, old_size);

            lightList = tmp_nodes;

            float[][] tmp_tx  = new float[new_size][];
            System.arraycopy(transformList, 0, tmp_tx, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_tx[i] = new float[16];

            transformList = tmp_tx;
        }
    }

    /**
     * Resize the transform stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    private final void resizeStack()
    {
        if((lastStack + 1) == transformStack.length)
        {
            int old_size = transformStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Matrix4f[] tmp_matrix = new Matrix4f[new_size];

            System.arraycopy(transformStack, 0, tmp_matrix, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_matrix[i] = new Matrix4f();

            transformStack = tmp_matrix;
        }
    }

    private void updateFrustum(RenderEnvironmentData envData)
    {
        // Create projection matrix
        float left = (float)viewFrustum[0];
        float right = (float)viewFrustum[1];
        float bottom = (float)viewFrustum[2];
        float top = (float)viewFrustum[3];
        float nearval = (float)viewFrustum[4];
        float farval = (float)viewFrustum[5];
        float x, y, z, w;
        float a, b, c, d;

        x = (2.0f * nearval) / (right - left);
        y = (2.0f * nearval) / (top - bottom);
        a = (right + left) / (right - left);
        b = (top + bottom) / (top - bottom);
        c = -(farval + nearval) / ( farval - nearval);
        d = -(2.0f * farval * nearval) / (farval - nearval);

        prjMatrix.m00 = x;
        prjMatrix.m01 = 0;
        prjMatrix.m02 = a;
        prjMatrix.m03 = 0;
        prjMatrix.m10 = 0;
        prjMatrix.m11 = y;
        prjMatrix.m12 = b;
        prjMatrix.m13 = 0;
        prjMatrix.m20 = 0;
        prjMatrix.m21 = 0;
        prjMatrix.m22 = c;
        prjMatrix.m23 = d;
        prjMatrix.m30 = 0;
        prjMatrix.m31 = 0;
        prjMatrix.m32 = -1;
        prjMatrix.m33 = 0;

        envData.projectionMatrix.set(prjMatrix);

        try
        {
            viewMatrix.invert();
        }
        catch(Exception e)
        {
        }

        viewMatrix.mul(prjMatrix,viewMatrix);

        // Put result into opengl format
        prjMatrix.m00 = viewMatrix.m00;
        prjMatrix.m01 = viewMatrix.m10;
        prjMatrix.m02 = viewMatrix.m20;
        prjMatrix.m03 = viewMatrix.m30;
        prjMatrix.m10 = viewMatrix.m01;
        prjMatrix.m11 = viewMatrix.m11;
        prjMatrix.m12 = viewMatrix.m21;
        prjMatrix.m13 = viewMatrix.m31;
        prjMatrix.m20 = viewMatrix.m02;
        prjMatrix.m21 = viewMatrix.m12;
        prjMatrix.m22 = viewMatrix.m22;
        prjMatrix.m23 = viewMatrix.m32;
        prjMatrix.m30 = viewMatrix.m03;
        prjMatrix.m31 = viewMatrix.m13;
        prjMatrix.m32 = viewMatrix.m23;
        prjMatrix.m33 = viewMatrix.m33;

        float t;
        /* Extract the numbers for the RIGHT plane */
        x = prjMatrix.m03 - prjMatrix.m00;
        y = prjMatrix.m13 - prjMatrix.m10;
        z = prjMatrix.m23 - prjMatrix.m20;
        w = prjMatrix.m33 - prjMatrix.m30;

        /* Normalize the result */
        t = 1.0f / (float) Math.sqrt(x * x + y * y + z * z );
        frustumPlanes[0].x = x * t;
        frustumPlanes[0].y = y * t;
        frustumPlanes[0].z = z * t;
        frustumPlanes[0].w = w * t;

        /* Extract the numbers for the LEFT plane */
        x = prjMatrix.m03 + prjMatrix.m00;
        y = prjMatrix.m13 + prjMatrix.m10;
        z = prjMatrix.m23 + prjMatrix.m20;
        w = prjMatrix.m33 + prjMatrix.m30;

        t = (float) Math.sqrt(x * x + y * y + z * z);
        frustumPlanes[1].x = x * t;
        frustumPlanes[1].y = y * t;
        frustumPlanes[1].z = z * t;
        frustumPlanes[1].w = w * t;

        /* Extract the BOTTOM plane */
        x = prjMatrix.m03 + prjMatrix.m01;
        y = prjMatrix.m13 + prjMatrix.m11;
        z = prjMatrix.m23 + prjMatrix.m21;
        w = prjMatrix.m33 + prjMatrix.m31;

        t = (float) Math.sqrt(x * x + y * y + z * z);
        frustumPlanes[2].x = x * t;
        frustumPlanes[2].y = y * t;
        frustumPlanes[2].z = z * t;
        frustumPlanes[2].w = w * t;

        /* Extract the TOP plane */
        x = prjMatrix.m03 - prjMatrix.m01;
        y = prjMatrix.m13 - prjMatrix.m11;
        z = prjMatrix.m23 - prjMatrix.m21;
        w = prjMatrix.m33 - prjMatrix.m31;

        t = (float) Math.sqrt(x * x + y * y + z * z);
        frustumPlanes[3].x = x * t;
        frustumPlanes[3].y = y * t;
        frustumPlanes[3].z = z * t;
        frustumPlanes[3].w = w * t;

        /* Extract the FAR plane */
        x = prjMatrix.m03 - prjMatrix.m02;
        y = prjMatrix.m13 - prjMatrix.m12;
        z = prjMatrix.m23 - prjMatrix.m22;
        w = prjMatrix.m33 - prjMatrix.m32;

        t = (float) Math.sqrt(x * x + y * y + z * z);
        frustumPlanes[4].x = x * t;
        frustumPlanes[4].y = y * t;
        frustumPlanes[4].z = z * t;
        frustumPlanes[4].w = w * t;

        /* Extract the NEAR plane */
        x = prjMatrix.m03 + prjMatrix.m02;
        y = prjMatrix.m13 + prjMatrix.m12;
        z = prjMatrix.m23 + prjMatrix.m22;
        w = prjMatrix.m33 + prjMatrix.m32;

        t = (float) Math.sqrt(x * x + y * y + z * z);
        frustumPlanes[5].x = x * t;
        frustumPlanes[5].y = y * t;
        frustumPlanes[5].z = z * t;
        frustumPlanes[5].w = w * t;
    }

    private String levelTab() {
        StringBuffer res = new StringBuffer();
        for(int i=0; i < level; i++) {
            res.append("  ");
        }

        return res.toString();
    }
}