/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.Arrays;
import java.util.HashMap;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Implementation of the sort stage that does everything - state sorting and
 * depth sorted transparency.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.10 $
 */
public class StateAndTransparencyDepthSortStage implements SortStage
{
    /** Message when the number of depth bits <= 0 */
    private static final String INVALID_DEPTH_MSG =
        "The depthBits must be greater than or equal to 1";

    /** The initial size of the instruction list */
    private static final int LIST_START_SIZE = 1;

    /** The initial size of the transparent list */
    private static final int TRANSPARENT_LIST_SIZE = 200;

    /** The size to grow the render list if needed */
    private static final int REALLOC_SIZE = 1024;

    /** The guess at the number of components in a shape */
    private static final int GUESS_NUM_COMPONENTS = 5;

    /** Default accuracy bits. Matches most common video cards */
    private static final int DEFAULT_DEPTH_BITS = 16;

    /**
     * The accuracy of the depth quantisation. This will be a power of two
     * value that is used to setup the depth list for input to the counting
     * sort.
     */
    private final int depthAccuracy;

    /** Receiver for the output of this stage */
    private SortedGeometryReceiver receiver;

    /** Output array for passing on to the receiver */
    private RenderInstructions[] commandList;

    /** Temporary array for holding the transparent objects */
    private CullOutputDetails[] transparentList;

    /** Temporary array for holding the transparent objects prior to sorting */
    private CullOutputDetails[] unsortedTransparentList;

    /** Temporary array for holding the opaque objects before sorting */
    private CullOutputDetails[] opaqueList;

    /** Temporary array for holding the transparent object depths */
    private float[] depthList1;

    /** Temporary array for holding the quantized depths */
    private int[] depthList2;

    /** The count list used during sorting */
    private int[] countList;

    /** Number of valid nodes in the node & operation lists */
    private int totalCommands;

    /** Temp array for reading texture units from the appearance */
    private TextureUnit[] textureTmp;

    /** A semi-unique ID counter used for assigning light IDs. May wrap */
    private int lastGlobalId;

    /** Map of the first occurance of a scene parent to it's render instruction */
    private HashMap instructionMap;

    /** Comparator instance for the state sorting */
    private StateSortComparator stateComparator;

    /** Temp matrix used to hold the local to vworld values */
    private Matrix4f modelMatrix;

    /** Temp matrix used to hold inverted camera matrix */
    private Matrix4f cameraMatrix;

    /** Temporary point location while working out the camera-space depth */
    private Point3f wkPoint;

    /** Temporary array for fetching the center position */
    private float[] center;

    // Listing of the current items, while we're setting up the state handling.

    /** Temp array for holdin new lights on the object being processed */
    private LightDetails[] newLights;

    /** Temp array for fetching lights objects from the HashSets below. */
    private Object[] lightTmp;

    /** Set of current lights that are being processed */
    private HashSet currentLights;

    /** Set for holding the different between the current and new sets */
    private HashSet oldLights;

    /** Set to hold onto lights that are common between last node and this */
    private HashSet keepLights;

    /** Mapping of the LightDetails instance to the global ID */
    private HashMap lightIdMap;

    /** The currently valid appearance */
    private Appearance currentAppearance;

    /** The currently valid material */
    private Material currentMaterial;

    /** The currently valid polygon attributes */
    private PolygonAttributes currentPolyAttr;

    /** The currently valid line attributes */
    private LineAttributes currentLineAttr;

    /** The currently valid point attributes */
    private PointAttributes currentPointAttr;

    /** The currently valid blend attributes */
    private BlendAttributes currentBlendAttr;

    /** The currently valid GLSlang shader*/
    private Shader currentShader;

    /** The currently valid vertex shader */
    private VertexShader currentVtxShader;

    /** The currently valid fragment shader */
    private FragmentShader currentFragShader;

    /** Listing of currently valid textures */
    private TextureUnit[] currentTextures;

    /** Number of valid textures currently active */
    private int numTextures;

    /** The requested size for the arrays */
    int req_size;

    /**
     * Create an empty sorting stage that assumes just a single renderable
     * output.
     */
    public StateAndTransparencyDepthSortStage()
    {
        this(LIST_START_SIZE, DEFAULT_DEPTH_BITS);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen. The number of depth bits is set to 16.
     *
     * @param numSurfaces The number of surfaces to start the internal lists
     *     sizes at.
     */
    public StateAndTransparencyDepthSortStage(int numSurfaces)
    {
        this(numSurfaces, DEFAULT_DEPTH_BITS);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     * @param numSurfaces The number of surfaces to start the internal lists
     *     sizes at.
     * @param depthBits The number of bits of depth precision to quantize the
     *     depth sorting to
     * @throws IllegalArgumentException Depth bits is zero or negative
     */
    public StateAndTransparencyDepthSortStage(int numSurfaces, int depthBits)
    {
        if(depthBits < 1)
            throw new IllegalArgumentException(INVALID_DEPTH_MSG);

        int depth = 1;

        for(int i = 1; i < depthBits; i++)
            depth <<= 1;

        depthAccuracy = depth;

        transparentList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];
        unsortedTransparentList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];
        opaqueList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];

        depthList1 = new float[TRANSPARENT_LIST_SIZE];
        depthList2 = new int[TRANSPARENT_LIST_SIZE];
        countList = new int[depthAccuracy];
        instructionMap = new HashMap();
        wkPoint = new Point3f();
        modelMatrix = new Matrix4f();
        cameraMatrix = new Matrix4f();
        center = new float[3];

        lastGlobalId = 0;

        commandList = new RenderInstructions[numSurfaces];

        for(int i = 0; i < numSurfaces; i++)
            commandList[i] = new RenderInstructions();

        instructionMap = new HashMap();
        stateComparator = new StateSortComparator();

        currentLights = new HashSet();
        keepLights = new HashSet();
        oldLights = new HashSet();
        lightIdMap = new HashMap();
        lightTmp = new Object[24];
        newLights = new LightDetails[32];

        // 32 is max textures that OpenGL can support.
        currentTextures = new TextureUnit[32];
        textureTmp = new TextureUnit[32];
    }

    /**
     * Sort the listing of nodes in the given array. Do not return until the
     * sort has been completed. If the sceneParent is the main scene graph and
     * not an offscreen texture, then that spot should be set to null in the list.
     *
     * @param data External rendering environment information
     * @param nodes List of processed nodes based on the scene they came from
     * @param numNodes Number of nodes in each scene
     * @param sceneParents Parent nodes that hold the subscene in the main scene graph
     * @param numScenes The number of valid scenes to process
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void sort(RenderEnvironmentData[] data,
                     CullOutputDetails[][] nodes,
                     int[] numNodes,
                     SceneGraphObject[][] sceneParents,
                     int numScenes,
                     int buffer)
    {
        // Buffer ID is not used currently.

        if(commandList.length < numScenes)
        {
            RenderInstructions[] tmp = new RenderInstructions[numScenes];

            System.arraycopy(commandList, 0, tmp, 0, commandList.length);

            for(int i = commandList.length; i < numScenes   ; i++)
                tmp[i] = new RenderInstructions();

            commandList = tmp;
        }

        // While doing these loops, we need to find the main scene and ensure
        // that it is the last item in the output list.
        int main_idx = 0;
        int curr_scene = 0;

        for(int i = 0; i < numScenes; i++)
        {
            if(sceneParents[i][0] == null)
            {
                main_idx = i;
                continue;
            }

            commandList[curr_scene].pbuffer =
                (OffscreenTextureSource)sceneParents[i][0];
            commandList[curr_scene].parentSource =
                (OffscreenTextureSource)sceneParents[i][1];
            commandList[curr_scene].renderData = data[i];

            RenderInstructions original =
                (RenderInstructions)instructionMap.get(sceneParents[i][0]);

            if(original != null)
            {
                commandList[curr_scene].copyOf = original;
            }
            else
            {
                commandList[curr_scene].copyOf = null;
                sortSingle(data[i].viewTransform,
                           nodes[i],
                           numNodes[i],
                           commandList[curr_scene]);

                instructionMap.put(sceneParents[i][0],
                                   commandList[curr_scene]);
            }

            curr_scene++;
        }

        commandList[curr_scene].pbuffer = null;
        commandList[curr_scene].renderData = data[main_idx];

        sortSingle(data[main_idx].viewTransform,
                   nodes[main_idx],
                   numNodes[main_idx],
                   commandList[curr_scene]);

        curr_scene++;

        if(receiver != null)
            receiver.sortedOutput(commandList, numScenes);

        instructionMap.clear();
    }


    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setSortedGeometryReceiver(SortedGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Sort a single set of nodes into the output details and place in the
     * provided RenderInstructions instance.
     *
     * @param instr Instruction instant to put the details into
     */
    private void sortSingle(Matrix4f camera,
                            CullOutputDetails[] nodes,
                            int numNodes,
                            RenderInstructions instr)
    {
        int new_size = 2 + numNodes * GUESS_NUM_COMPONENTS;

        if (new_size > req_size)
            req_size = new_size;

        if(transparentList.length < req_size)
        {
            transparentList = new CullOutputDetails[req_size];

            unsortedTransparentList = new CullOutputDetails[req_size];

            depthList1 = new float[req_size];
            depthList2 = new int[req_size];

            opaqueList = new CullOutputDetails[req_size];
        }

        int idx = 0;
        int trans = 0;
        int opaque = 0;

        for(int i = 0; i < numNodes; i++)
        {
            // First check the geometry and so forth for something that
            // has an alpha channel. Need to check everything - geometry
            // colour values, textures and materials. Anything that says it
            // may have alpha is passed off onto the transparent list.
            // Now process the node type.
            if(nodes[i].renderable instanceof Shape3D)
            {
                Shape3D shape = (Shape3D)nodes[i].renderable;

                Appearance app = shape.getAppearance();

                if(app != null)
                {
                    // Anything that has explicit blend ops is not counted in
                    // the normal transparency sorting. We'll dump it back into
                    // the non-transparent bucket, not transparent.
                    BlendAttributes blender = app.getBlendAttributes();
                    if(blender != null)
                    {
                        unsortedTransparentList[trans++] = nodes[i];
                        continue;
                    }

                    Material mat = app.getMaterial();

                    if(mat != null && mat.getTransparency() != 1)
                    {
                        unsortedTransparentList[trans++] = nodes[i];
                        continue;
                    }

                    int num_tex = app.numTextureUnits();

                    if(num_tex != 0)
                    {
                        boolean alpha_found = false;

                        app.getTextureUnits(textureTmp);
                        for(int j = 0; j < num_tex; j++)
                        {
                            Texture tex = textureTmp[j].getTexture();
                            if(tex == null)
                                continue;

                            // Any of the following formats, this is is added
                            // to the transparency collection. Break out of the
                            // inner loop by messing with the loop index.
                            switch(tex.getFormat())
                            {
                                case Texture.FORMAT_ALPHA:
                                case Texture.FORMAT_INTENSITY_ALPHA:
                                case Texture.FORMAT_RGBA:
                                    unsortedTransparentList[trans++] = nodes[i];
                                    alpha_found = true;
                                    j = num_tex;
                                    break;

                            }
                        }

                        if(alpha_found)
                            continue;
                    }
                }

                Geometry geom = shape.getGeometry();
                if(geom.hasTransparency())
                {
                    req_size++;
                    unsortedTransparentList[trans++] = nodes[i];
                    continue;
                }
            }
            else
            {
                System.out.println("Non-shape node in sorter " +
                                   nodes[i].renderable);
                continue;
            }

            opaqueList[opaque++] = nodes[i];
//            idx = appendOutputShape(idx, nodes[i], instr);
        }

        Arrays.sort(opaqueList, 0, opaque, stateComparator);


        if(instr.renderList.length < req_size)
        {
            if (req_size - instr.renderList.length < REALLOC_SIZE)
                req_size = instr.renderList.length + REALLOC_SIZE;

            RenderDetails[] tmp = new RenderDetails[req_size];

            System.arraycopy(instr.renderList, 0, tmp, 0, instr.renderList.length);

            for(int i = instr.renderList.length; i < req_size; i++)
                tmp[i] = new RenderDetails();

            instr.renderList = tmp;
            instr.renderOps = new int[req_size];
        }

        int start = 0;
        boolean done = false;

        while(!done)
        {
            try {
                for(int i = start; i < opaque; i++)
                {
                    idx = appendObject(opaqueList[i], instr, idx);
                }

                done = true;
            } catch(ArrayIndexOutOfBoundsException be)
            {
                start = idx;

                realloc(instr);
            }
        }

        done = false;

        while(!done)
        {
            try {
                idx = cleanupObjects(instr, idx);

                done = true;
            } catch(ArrayIndexOutOfBoundsException be)
            {
                realloc(instr);
            }
        }

        new_size = idx + 1 + trans << 1;
        if (new_size > req_size)
            req_size = new_size;

        if(trans > 0 && instr.renderList.length < req_size)
        {
            realloc(instr);
        }

        // Now process the transparency listing. If there is no transparent
        // objects, then exit now.
        switch(trans)
        {
            case 0:
                instr.numValid = idx;
                return;

            case 1:
                instr.renderOps[idx++] = RenderOp.START_TRANSPARENT;

                idx = appendOutputShape(idx,
                                        unsortedTransparentList[0],
                                        instr);

                // clear it from the list
                unsortedTransparentList[0] = null;

                instr.renderOps[idx++] = RenderOp.STOP_TRANSPARENT;
                instr.numValid = idx;
                return;
        }

        // Time to work out the distance values.
        float max_depth = Float.NEGATIVE_INFINITY;
        float min_depth = Float.POSITIVE_INFINITY;

        cameraMatrix.invert(camera);

        for(int i = 0; i < trans; i++)
        {
            // transform the center of the bounds first to world space and then
            // into camera space to work out the depth.
            modelMatrix.set(unsortedTransparentList[i].transform);
            Shape3D shape = (Shape3D)unsortedTransparentList[i].renderable;
            BoundingVolume bounds = shape.getBounds();
            bounds.getCenter(center);

            wkPoint.x = center[0];
            wkPoint.y = center[1];
            wkPoint.z = center[2];

            modelMatrix.transform(wkPoint);
            cameraMatrix.transform(wkPoint);

            depthList1[i] = wkPoint.z;

            if(max_depth < wkPoint.z)
                max_depth = wkPoint.z;

            if(min_depth > wkPoint.z)
                min_depth = wkPoint.z;
        }


        // Now work out the range of values available and how to quantize those
        // to the given accuracy. Take 1 from the depth accuracy so that the max
        // depth will always end up at one less than the final array size - to
        // avoid AAOB exceptions.
        float depth_range = max_depth - min_depth;
        float quanta = (depthAccuracy - 1)/ depth_range;

        // Change all the values from floats to quantised ints
        for(int i = 0; i < trans; i++)
            depthList2[i] = (int)((depthList1[i] - min_depth) * quanta);

        // Now perform the counting sort on them. Initialise all the counts to
        // 0 again
        for(int i = 0; i < depthAccuracy; i++)
            countList[i] = 0;

        for(int j = 0; j < trans; j++)
            countList[depthList2[j]]++;

        // countList now contains the number of elements equal to i
        for(int i = 1; i < depthAccuracy; i++)
            countList[i] += countList[i - 1];

        for(int j = trans; --j >= 0; )
        {
            int dl = depthList2[j];
            transparentList[countList[dl] - 1] = unsortedTransparentList[j];
            countList[dl]--;
        }

        instr.renderOps[idx++] = RenderOp.START_TRANSPARENT;


        done = false;
        start = 0;

        while(!done)
        {
            try {
                for(int i = start; i < trans; i++)
                {
                    idx = appendOutputShape(idx, transparentList[i], instr);

                    // clear it from the list
                    transparentList[i] = null;
                }

                done = true;
            } catch(ArrayIndexOutOfBoundsException be)
            {
                start = idx;

                realloc(instr);
            }
        }

        instr.renderOps[idx++] = RenderOp.STOP_TRANSPARENT;
        instr.numValid = idx;
    }

    /**
     * Convenience method to add a Shape3D to the list at the given index point.
     *
     * @param idx The starting place in the instruction list
     * @param node The node instance to add
     * @param instr List of instructions to add the node to
     * @return The update idx value taking into account the new nodes
     */
    private int appendOutputShape(int idx,
                                 CullOutputDetails node,
                                 RenderInstructions instr)
    {
        if(node.numLights != 0)
        {
            for(int j = 0; j < node.numLights; j++)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_LIGHT;
                idx++;
            }
        }

        // Now process the node type.
        // For this simple one, just use the basic render command.
        // Something more complex would pull the shape apart and do
        // state/depth/transparency sorting in this section.

        instr.renderList[idx].renderable = node.renderable;

        instr.renderList[idx].transform[0] = node.transform.m00;
        instr.renderList[idx].transform[1] = node.transform.m10;
        instr.renderList[idx].transform[2] = node.transform.m20;
        instr.renderList[idx].transform[3] = node.transform.m30;

        instr.renderList[idx].transform[4] = node.transform.m01;
        instr.renderList[idx].transform[5] = node.transform.m11;
        instr.renderList[idx].transform[6] = node.transform.m21;
        instr.renderList[idx].transform[7] = node.transform.m31;

        instr.renderList[idx].transform[8] = node.transform.m02;
        instr.renderList[idx].transform[9] = node.transform.m12;
        instr.renderList[idx].transform[10] = node.transform.m22;
        instr.renderList[idx].transform[11] = node.transform.m32;

        instr.renderList[idx].transform[12] = node.transform.m03;
        instr.renderList[idx].transform[13] = node.transform.m13;
        instr.renderList[idx].transform[14] = node.transform.m23;
        instr.renderList[idx].transform[15] = node.transform.m33;

        instr.renderOps[idx] = RenderOp.START_RENDER;
        idx++;

        instr.renderList[idx].renderable = node.renderable;
        instr.renderOps[idx] = RenderOp.STOP_RENDER;
        idx++;

        if(node.numLights != 0)
        {
            for(int j = node.numLights - 1; j >= 0; j--)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();
                instr.renderList[idx].id = lastGlobalId - (1 + j);

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;
            }
        }

        return idx;
    }

    /**
     * Append a single object onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int appendObject(CullOutputDetails node,
                             RenderInstructions instr,
                             int offset)
    {
        int idx = offset;

        // First drop all the lights into the queue and turn the state off on
        // the no-longer used ones.
        if(node.numLights != 0)
        {
            int new_count = 0;

            for(int i = 0; i < node.numLights; i++)
            {
                LightDetails ld = node.lights[i];

                if(currentLights.contains(ld))
                    keepLights.add(ld);
                else
                    newLights[new_count++] = ld;
            }

            currentLights.retainAll(keepLights, oldLights);
            keepLights.clear();

            // Unmap the old lights.
            int num_old = oldLights.size();
            if(num_old != 0)
            {
                if(num_old > lightTmp.length)
                    lightTmp = new Object[num_old];

                oldLights.toArray(lightTmp);
                oldLights.clear();

                for(int i = 0; i < num_old; i++)
                {
                    LightDetails ld = (LightDetails)lightTmp[i];
                    Integer l_id = (Integer)lightIdMap.remove(ld);

                    instr.renderList[idx].renderable = ld.getLight();
                    instr.renderList[idx].id = l_id.intValue();

                    // Don't need the transform for the light stopping. Save
                    // CPU cycles by not copying it.

                    instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                    idx++;

                    lightTmp[i] = null;
                }
            }

            // Map the new lights.
            if(new_count != 0)
            {
                for(int i = 0; i < new_count; i++)
                {
                    LightDetails ld = newLights[i];
                    currentLights.add(ld);

                    instr.renderList[idx].renderable = ld.getLight();

                    System.arraycopy(ld.getTransform(),
                                     0,
                                     instr.renderList[idx].transform,
                                     0,
                                     16);

                    lightIdMap.put(ld, new Integer(lastGlobalId));
                    instr.renderList[idx].id = lastGlobalId++;
                    instr.renderOps[idx] = RenderOp.START_LIGHT;
                    idx++;

                    newLights[i] = null;
                }
            }
        }
        // Finally, map the shape instances
        if(node.renderable instanceof Shape3D)
        {
            Shape3D shape = (Shape3D)node.renderable;
            Appearance app = shape.getAppearance();
            Geometry geom = shape.getGeometry();

            if(currentAppearance != app)
            {
                // If the new state is now null, but before it wasn't then
                // we need to clear all the current visual state. Otherwise
                // we have to change the state from the existing one and
                // setup the new state.
                if(app == null)
                {
                    idx = cleanupVisuals(instr, idx);
                }
                else
                {
                    idx = updateMaterial(app, instr, idx);
                    idx = updatePolyAttribs(app, instr, idx);
                    idx = updateLineAttribs(app, instr, idx);
                    idx = updatePointAttribs(app, instr, idx);
                    idx = updateBlendAttribs(app, instr, idx);
                    idx = updateTextures(app, instr, idx);
                }

                currentAppearance = app;
            }

            instr.renderList[idx].renderable = geom;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.START_GEOMETRY;
            idx++;

            instr.renderList[idx].renderable = geom;
            instr.renderOps[idx] = RenderOp.STOP_GEOMETRY;
            idx++;
        }
        else
        {
            System.out.println("Non-shape node in sorter " + node.renderable);
        }

        return idx;
    }

    /**
     * Check on the material setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateMaterial(Appearance app,
                               RenderInstructions instr,
                               int offset)
    {
        int idx = offset;

        Material mat = app.getMaterial();
        if(mat != currentMaterial)
        {
            if(mat != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!mat.equals(currentMaterial))
                {
                    // Clear the old state if it is set
                    if(currentMaterial != null)
                    {
                        instr.renderList[idx].renderable = currentMaterial;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = mat;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentMaterial = mat;
            }
            else
            {
                // The new material is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentMaterial;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentMaterial = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updatePolyAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        PolygonAttributes attr = app.getPolygonAttributes();
        if(attr != currentPolyAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPolyAttr))
                {
                    // Clear the old state if it is set
                    if(currentPolyAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPolyAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPolyAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPolyAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPolyAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the line attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateLineAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        LineAttributes attr = app.getLineAttributes();
        if(attr != currentLineAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentLineAttr))
                {
                    // Clear the old state if it is set
                    if(currentLineAttr != null)
                    {
                        instr.renderList[idx].renderable = currentLineAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentLineAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentLineAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentLineAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updatePointAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        PointAttributes attr = app.getPointAttributes();
        if(attr != currentPointAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPointAttr))
                {
                    // Clear the old state if it is set
                    if(currentPointAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPointAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPointAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPointAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPointAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the blending attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateBlendAttribs(Appearance app,
                                   RenderInstructions instr,
                                   int offset)
    {
        int idx = offset;

        BlendAttributes attr = app.getBlendAttributes();
        if(attr != currentBlendAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentBlendAttr))
                {
                    // Clear the old state if it is set
                    if(currentBlendAttr != null)
                    {
                        instr.renderList[idx].renderable = currentBlendAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentBlendAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentBlendAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentBlendAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the texture listing setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateTextures(Appearance app,
                               RenderInstructions instr,
                               int offset)
    {
        int idx = offset;

        int num_tex = app.numTextureUnits();

        if(textureTmp.length < num_tex)
            textureTmp = new TextureUnit[num_tex];

        app.getTextureUnits(textureTmp);

        if(num_tex != numTextures)
        {
            if(num_tex == 0)
            {
                // Clear the old textures that were set
                for(int i = 0; i < numTextures; i++)
                {
                    instr.renderList[idx].renderable = currentTextures[i];
                    instr.renderOps[idx] = RenderOp.STOP_STATE;
                    idx++;

                    currentTextures[i] = null;
                }
            }
            else if(numTextures == 0)
            {
                // There were no old textures, so just list the new ones
                for(int i = 0; i < num_tex; i++)
                {
                    instr.renderList[idx].renderable = textureTmp[i];
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;

                    currentTextures[i] = textureTmp[i];
                }
            }
            else
            {
                int smaller = (numTextures < num_tex) ? numTextures : num_tex;

                for(int i = 0; i < smaller; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderOps[idx] = RenderOp.START_STATE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                    }
                }

                // Clean up what is left or add the new stuff
                if(numTextures < num_tex)
                {
                    for(int i = smaller; i < num_tex; i++)
                    {
                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderOps[idx] = RenderOp.START_STATE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                    }
                }
                else
                {
                    for(int i = smaller; i < numTextures; i++)
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;

                        currentTextures[i] = null;
                    }
                }
            }

            numTextures = num_tex;
        }
        else if(num_tex != 0)
        {
            // same number of texture units.
            // Special case the most common situation.
            if(num_tex == 1)
            {
                if(!textureTmp[0].equals(currentTextures[0]))
                {
                    instr.renderList[idx].renderable = currentTextures[0];
                    instr.renderOps[idx] = RenderOp.STOP_STATE;
                    idx++;

                    instr.renderList[idx].renderable = textureTmp[0];
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;

                    currentTextures[0] = textureTmp[0];
                    textureTmp[0] = null;
                }
            }
            else
            {
                for(int i = 0; i < numTextures; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderOps[idx] = RenderOp.START_STATE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }
            }
        }

        return idx;
    }

    /**
     * At the end of the frame, we need to make sure that we've turned off all
     * the lights and current rendering state that is left. Do that now.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int cleanupObjects(RenderInstructions instr, int offset)
    {
        int idx = offset;

        int num_lights = currentLights.size();
        if(num_lights != 0)
        {
            if(num_lights > lightTmp.length)
                lightTmp = new Object[num_lights];

            currentLights.toArray(lightTmp);
            currentLights.clear();

            for(int i = 0; i < num_lights; i++)
            {
                LightDetails ld = (LightDetails)lightTmp[i];
                Integer l_id = (Integer)lightIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getLight();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;

                lightTmp[i] = null;
            }
        }

        return cleanupVisuals(instr, idx);
    }

    /**
     * Convenience method to clear all the current visual information.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int cleanupVisuals(RenderInstructions instr, int offset)
    {
        int idx = offset;

        if(currentMaterial != null)
        {
            instr.renderList[idx].renderable = currentMaterial;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentMaterial = null;
        }

        if(currentShader != null)
        {
            instr.renderList[idx].renderable = currentShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentShader = null;
        }

        if(currentVtxShader != null)
        {
            instr.renderList[idx].renderable = currentVtxShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentVtxShader = null;
        }

        if(currentFragShader != null)
        {
            instr.renderList[idx].renderable = currentFragShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentFragShader = null;
        }

        if(currentPolyAttr != null)
        {
            instr.renderList[idx].renderable = currentPolyAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPolyAttr = null;
        }

        if(currentLineAttr != null)
        {
            instr.renderList[idx].renderable = currentLineAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentLineAttr = null;
        }

        if(currentPointAttr != null)
        {
            instr.renderList[idx].renderable = currentPointAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPointAttr = null;
        }

        if(currentBlendAttr != null)
        {
            instr.renderList[idx].renderable = currentBlendAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentBlendAttr = null;
        }

        for(int i = 0; i < numTextures; i++)
        {
            instr.renderList[idx].renderable = currentTextures[i];
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentTextures[i] = null;
        }

        numTextures = 0;
        currentAppearance = null;

        return idx;
    }

    /**
     * Reallocate the renderList and renderOps arrays.
     * Grows by a minimum of REALLOC_SIZE.
     *
     * @param instr The instructions to resize.
     */
    private void realloc(RenderInstructions instr)
    {
        if (req_size - instr.renderList.length < REALLOC_SIZE)
            req_size += REALLOC_SIZE;

        RenderDetails[] tmp = new RenderDetails[req_size];
        int[] roTmp = new int[req_size];

        System.arraycopy(instr.renderList, 0, tmp, 0, instr.renderList.length);
        System.arraycopy(instr.renderOps, 0, roTmp, 0, instr.renderOps.length);

        for(int i = instr.renderList.length; i < req_size; i++)
            tmp[i] = new RenderDetails();

        instr.renderList = tmp;
        instr.renderOps = roTmp;
    }
}